import abc
import numpy  # help IDEs with type-hinting inside docstrings
import numpy as np
import six
import textwrap
import warnings
import itertools
import matplotlib as mpl
import os
from collections.abc import Iterable

from .container import DataContainerBase
from .format import ParameterFormatter
from ..multi.fit import MultiFit
from ...config import kc, ConfigError, kafe2_rc

from collections import OrderedDict
from matplotlib import pyplot as plt
from matplotlib import gridspec as gs
from matplotlib.legend_handler import HandlerBase
from matplotlib import rc_context

__all__ = ["PlotAdapterBase", "Plot", "kc_plot_style"]


def kc_plot_style(data_type, subplot_key, property_key):
    try:
        # try to find plot style-related configuration entry
        return kc('fit', 'plot', 'style', data_type, subplot_key, property_key)
    except ConfigError:
        # if not available, do lookup for the default data type
        return kc('fit', 'plot', 'style', 'default', subplot_key, property_key)


class Cycler(object):
    # TODO: handle_mismatching_lengths in ['repeat', 'repeat_last', 'reflect']
    def __init__(self, *args):
        self._props = []
        self._modulo = 1

        # read in properties and check
        _pv_sizes = []
        _processed_names = set()
        for _i, _content in enumerate(args):
            _prop_size_i = None
            _prop_dict_i = dict()
            for _prop_name, _prop_vals in six.iteritems(_content):
                # for the time being: 'refuse' any mismatching property value lengths
                if _prop_size_i is None:
                    _prop_size_i = len(_prop_vals)
                else:
                    if len(_prop_vals) != _prop_size_i:
                        raise ValueError("Cannot cycle properties with mismatching value "
                                         "sequence lengths!")
                if _prop_name in _processed_names:
                    raise ValueError(
                        "Cycle already contains a property named '%s'!" % (_prop_name,))
                _prop_dict_i[_prop_name] = tuple(_prop_vals)
                _processed_names.add(_prop_name)
            _pv_sizes.append(_prop_size_i)
            self._modulo *= _prop_size_i
            self._props.append(_prop_dict_i)
        self._dim = len(self._props)

        self._prop_val_sizes = np.array(_pv_sizes, dtype=int)
        self._counter_divisors = np.ones_like(self._prop_val_sizes, dtype=int)
        for i in range(1, self._dim):
            self._counter_divisors[i] = self._counter_divisors[i-1] * self._prop_val_sizes[i-1]
        self._cycle_counter = 0

    # public properties

    @property
    def modulo(self):
        return self._modulo

    # public methods

    def get(self, cycle_position):
        _prop_positions = [(cycle_position//self._counter_divisors[i]) % self._prop_val_sizes[i]
                           for i in six.moves.range(self._dim)]
        _ps = {}
        for _i, _content in enumerate(self._props):
            for (_name, _values) in six.iteritems(_content):
                _ps[_name] = _values[_prop_positions[_i]]
        return _ps

    def get_next(self):
        _ps = self.get(self._cycle_counter)
        self._cycle_counter += 1
        return _ps

    def reset(self):
        # TODO: test
        self._cycle_counter = 0

    def combine(self, other_cycler):
        # TODO: test
        # TODO: more control over combination
        _args = self._props + other_cycler._props
        return Cycler(*_args)

    def subset_cycler(self, properties):
        # TODO: test
        _args = []
        for _i, _content in enumerate(self._props):
            _tmp_dict = {}
            for (_name, _values) in six.iteritems(_content):
                if _name in properties:
                    _tmp_dict[_name] = _values
            _args.append(_tmp_dict)
        return Cycler(*_args)


class DummyLegendHandler(HandlerBase):
    """Dummy legend handler (nothing is drawn)"""
    def legend_artist(self, *args, **kwargs):
        return None


@six.add_metaclass(abc.ABCMeta)
class PlotAdapterBase:
    """This is a purely abstract class implementing the minimal interface required by all
    types of plot adapters.

    A :py:obj:`PlotAdapter` object can be constructed for a :py:obj:`Fit` object of the
    corresponding type.
    Its main purpose is to provide an interface for accessing data stored in the
    :py:obj:`Fit` object, for the purposes of plotting.
    Most importantly, it provides methods to call the relevant :py:mod:`matplotlib` methods
    for plotting the data, model (and other information, depending on the fit type),
    and constructs the arrays required by these routines in a meaningful way.

    Classes derived from :py:class:`PlotAdapterBase` must at the very least contain
    properties for constructing the *x* and *y* point arrays for both the
    data and the fitted model, as well as methods calling the :py:mod:`matplotlib` routines
    doing the actual plotting.
    """

    PLOT_STYLE_CONFIG_DATA_TYPE = 'default'

    PLOT_SUBPLOT_TYPES = OrderedDict(
        data=dict(
            plot_adapter_method='plot_data',
            target_axes='main',
            container_valid=True,
        ),
        model=dict(
            plot_adapter_method='plot_model',
            target_axes='main',
        ),
        ratio=dict(
            plot_style_as='data',
            plot_adapter_method='plot_ratio',
            target_axes='ratio',
        ),
        residual=dict(
            plot_style_as='data',
            plot_adapter_method='plot_residual',
            target_axes='residual',
        ),
    )

    AVAILABLE_X_SCALES = ('linear',)
    AVAILABLE_Y_SCALES = ('linear', 'log')

    def __init__(self, fit_object, from_container=False):
        """Construct a :py:obj:`PlotAdapter` for a :py:obj:`Fit` object:

        :param fit_object: An object derived from :py:obj:`~.FitBase`
        :type fit_object: kafe2.fit._base.FitBase
        :param from_container: Whether fit_object was created ad-hoc from just a data container.
        :type from_container: bool
        """
        self._fit = fit_object
        self._from_container = from_container
        self._x_range = None
        self._y_range = None
        self._x_scale = 'linear'
        self._y_scale = 'linear'
        _axes = ('x', 'y')
        _axis_labels = list()
        for i, axis in enumerate(_axes):
            label = self._fit.data_container.axis_labels[i]  # load from data_container
            if label is None and axis == 'x' and type(fit_object).__name__ == "XYFit":
                _latex_name = fit_object.model_function.formatter.arg_formatters[0].latex_name
                label = f"${_latex_name}$"
            if label is None:  # fallback to default
                label = kc_plot_style(self.PLOT_STYLE_CONFIG_DATA_TYPE, 'axis_labels', axis)
            if label == '__del__':  # set axis label to None for special string __del__
                label = None
            _axis_labels.append(label)
        self._axis_labels = _axis_labels
        _x_ticks = fit_object._get_default_x_ticks()
        self._ticks = [_x_ticks, []]

        # specification of subplots for which this adapter provided plot routines
        self._subplots = None
        self._get_subplots()

        # set labels if present and according subplots are available
        self._set_plot_labels()

    def _get_subplots(self):
        """Create dictionary containing all subplot specifications.

        :return: An ordered dictionary containing all information about the available plot methods.
        :rtype: OrderedDict
        """
        if not self._subplots:
            # create subplot dict
            self._subplots = OrderedDict()
            for _pt, _pt_spec in six.iteritems(self.PLOT_SUBPLOT_TYPES):
                try:
                    # replace plot adapter method string with real method
                    self._subplots[_pt] = dict(
                        _pt_spec,
                        plot_adapter_method=getattr(self, _pt_spec['plot_adapter_method']),
                    )
                except KeyError as _e:
                    raise ValueError(
                        "Invalid subplot configuration: missing "
                        "key `plot_adapter_method` for subplot type '{}' "
                        "in PLOT_SUBPLOT_TYPES in {}!".format(
                            _pt,
                            self.__class__
                        )
                    ) from _e
                except AttributeError as _e:
                    raise TypeError(
                        "Cannot handle plot of type '{}': "
                        "cannot find corresponding plot method "
                        "'{}' in {}!".format(
                            _pt,
                            _pt_spec['plot_adapter_method'],
                            self.__class__
                        )
                    ) from _e
                self._subplots[_pt].setdefault('plot_method_keywords', {})

        return self._subplots

    def _set_plot_labels(self):
        """Obtain the data and model labels from data and model container and set them
        accordingly.
        """
        if self._fit.data_container.label is not None:
            try:
                self.update_plot_kwargs('data', dict(label=self._fit.data_container.label))
            except ValueError:
                pass  # no data present
        if self._fit.model_label is not None:
            for plot_model_name in ('model', 'model_line'):
                try:
                    self.update_plot_kwargs(plot_model_name, dict(label=self._fit.model_label))
                except ValueError:
                    pass  # no model plot function available
            _model_error_name = kc('fit', 'plot', 'error_label') % dict(model_label=self._fit.model_label) \
                if self._fit.model_label != '__del__' else '__del__'
            try:
                self.update_plot_kwargs('model_error_band', dict(label=_model_error_name))
            except ValueError:
                pass  # no error band available

    def _get_subplot_kwargs(self, plot_index, plot_type):
        """Resolve the keyword arguments passed to the plot method."""
        _subplots = self._get_subplots()

        try:
            _subplot = _subplots[plot_type]
        except KeyError as exc:
            raise ValueError(
                f"Unknown subplot: {plot_type} . Available: {list(_subplots.keys())}") from exc
        _explicit_kwargs = _subplot.get('plot_method_keywords', {})

        # get static kwargs
        _plot_style_as = _subplots[plot_type].get('plot_style_as', plot_type)

        # retrieve default plot keywords from style config
        _kwargs = dict(kc_plot_style(self.PLOT_STYLE_CONFIG_DATA_TYPE, _plot_style_as, 'plot_kwargs'))

        # initialize property cycler from style config and commit keywords
        _prop_cycler_args = kc_plot_style(self.PLOT_STYLE_CONFIG_DATA_TYPE, _plot_style_as, 'property_cycler')
        _prop_cycler = Cycler(*_prop_cycler_args)
        _kwargs.update(**_prop_cycler.get(plot_index))

        # override keywords with one explicitly set via the API
        _kwargs.update(**_explicit_kwargs)

        # remove keywords set to the special value '__del__'
        _kwargs = {
            _k : _v
            for _k, _v in six.iteritems(_kwargs)
            if _v != '__del__'
        }

        # apply interpolation to legend labels
        _label = _kwargs.pop('label', None)
        if _label:
            _kwargs['label'] = _label % dict(subplot_id=plot_index,
                                             plot_type=plot_type)

        # calculate zorder if not explicitly given
        _n_defined_plot_types = len(_subplots)
        if 'zorder' not in _kwargs:
            _kwargs['zorder'] = plot_index * _n_defined_plot_types + list(_subplots).index(plot_type)

        _container_valid = _subplots[plot_type].get("container_valid", None)
        if _container_valid is not None:
            _kwargs["container_valid"] = _container_valid

        return _kwargs

    def _get_total_error(self, error_contributions):
        _total_err = np.zeros_like(self.data_y)
        for _ec in error_contributions:
            _ec = _ec.lower()
            if _ec not in ('data', 'model'):
                raise ValueError(
                    "Unknown error contribution specification '{}': "
                    "expecting 'data' or 'model'".format(_ec))
            _total_err += getattr(self, _ec + '_yerr') ** 2
            _total_err += self._fit._cost_function.get_uncertainty_gaussian_approximation(
                getattr(self, _ec + '_y')) ** 2

        _total_err = np.sqrt(_total_err)

        if np.all(_total_err == 0):
            return None
        return _total_err

    # -- public API

    def call_plot_method(self, plot_type, target_axes, **kwargs):
        """Call the registered plot method for `plot_type`.

        :param plot_type: key identifying a registered plot type for this :py:obj:`PlotAdapter`
        :type plot_type: str
        :param target_axes: axes to plot to
        :type target_axes: `matplotlib.Axes` object
        :param kwargs: keyword arguments to pass to the plot method
        :type kwargs: dict
        :return: return value of the plot method
        """
        _subplots = self._get_subplots()

        if plot_type not in _subplots:
            raise ValueError(
                "Cannot call plot method: unknown plot type '{}'! "
                "Expecting one of: {!r}".format(
                    plot_type, list(_subplots)))

        _callable = _subplots[plot_type].get('plot_adapter_method', None)

        if _callable is None:
            raise ValueError(
                "Cannot call plot method: missing key `plot_adapter_method` "
                "in configuration for plot type '{}'!".format(
                    plot_type))
        if not callable(_callable):
            raise ValueError(
                "Cannot call plot method: registered `plot_adapter_method` "
                "with type {!r} in configuration for plot type '{}' "
                "is not callable!".format(
                    type(_callable), plot_type))

        if kwargs.pop('hide', False):
            return None

        if not kwargs.pop('container_valid', False) and self._from_container:
            return

        return _callable(
            target_axes=target_axes,
            **kwargs
        )

    def update_plot_kwargs(self, plot_type, plot_kwargs):
        """Update the value of keyword arguments `plot_kwargs` to be passed
        to the plot method for for `plot_type`.

        If a keyword argument should be removed, the value of the keyword
        in `plot_kwargs` can be set to the special value ``'__del__'``.
        To indicate that the default value should be used, the special
        value ``'__default__'`` can be set as a value.

        :param plot_type: key identifying a registered plot type for this :py:obj:`PlotAdapter`
        :type plot_type: str
        :param plot_kwargs: dictionary containing keywords arguments to override
        :type plot_kwargs: dict
        """
        _subplots = self._get_subplots()

        if plot_type not in _subplots:
            raise ValueError(
                "Cannot set custom plot keyword arguments "
                "for plot type '{}': no plot with this type defined "
                "for this adapter!".format(plot_type))

        # remove keys set to '__default__': the defaults will be substituted in
        # then _get_subplot_kwargs() is called. The '__del__' special value is
        # handled by _get_subplot_kwargs()
        _keys_to_delete = [_k for _k, _v in six.iteritems(plot_kwargs) if _v == '__default__']

        _config_dict = _subplots[plot_type].setdefault('plot_method_keywords', {})
        _config_dict.update(plot_kwargs)

        for _key in _keys_to_delete:
            try:
                del _config_dict[_key]
            except KeyError:
                pass  # ok, key not present

    # -- properties

    @property
    @abc.abstractmethod
    def data_x(self):
        """The *x* coordinates of the data (used by :py:meth:`~plot_data`).

        :rtype: numpy.ndarray
        """
        pass

    @property
    @abc.abstractmethod
    def data_y(self):
        """The *y* coordinates of the data (used by :py:meth:`~plot_data`).

        :rtype: numpy.ndarray
        """
        pass

    @property
    @abc.abstractmethod
    def data_xerr(self):
        """The magnitude of the data *x* error bars (used by :py:meth:`~plot_data`).

        :rtype: numpy.ndarray
        """
        pass

    @property
    @abc.abstractmethod
    def data_yerr(self):
        """The magnitude of the data *y* error bars (used by :py:meth:`~plot_data`).

        :rtype: numpy.ndarray
        """
        pass

    @property
    @abc.abstractmethod
    def model_x(self):
        """The *x* coordinates of the model (used by :py:meth:`~plot_model`).

        :rtype: numpy.ndarray
        """
        pass

    @property
    @abc.abstractmethod
    def model_y(self):
        """The *y* coordinates of the model (used by :py:meth:`~plot_model`).

        :rtype: numpy.ndarray
        """
        pass

    @property
    @abc.abstractmethod
    def model_xerr(self):
        """The magnitude of the model *x* error bars (used by :py:meth:`~plot_model`).

        :rtype: numpy.ndarray
        """
        pass

    @property
    @abc.abstractmethod
    def model_yerr(self):
        """The magnitude of the model *y* error bars (used by :py:meth:`~plot_model`).

        :rtype: numpy.ndarray
        """
        pass

    @property
    def x_range(self):
        """The *x* axis plot range.

        :rtype: tuple[float, float]
        """
        return self._x_range

    @x_range.setter
    def x_range(self, x_range):
        self._x_range = tuple(x_range)

    @property
    def y_range(self):
        """The *y* axis plot range.

        :rtype: tuple[float, float]
        """
        return self._y_range

    @y_range.setter
    def y_range(self, y_range):
        self._y_range = tuple(y_range)

    @property
    def x_scale(self):
        """The *x* axis scale. Available scales are given in :py:const:`~.AVAILABLE_X_SCALES`

        :rtype: str
        """
        return self._x_scale

    @x_scale.setter
    def x_scale(self, scale):
        if scale not in self.AVAILABLE_X_SCALES:
            raise TypeError("x_scale {} is not supported for this type of fit, "
                            "use one of {}".format(scale, self.AVAILABLE_X_SCALES))
        self._x_scale = scale

    @property
    def y_scale(self):
        """The *y* axis scale. Available scales are given in :py:const:`~.AVAILABLE_Y_SCALES`

        :rtype: str
        """
        return self._y_scale

    @y_scale.setter
    def y_scale(self, scale):
        if scale not in self.AVAILABLE_Y_SCALES:
            raise TypeError("y_scale {} is not supported for this type of fit, "
                            "use one of {}".format(scale, self.AVAILABLE_Y_SCALES))
        self._y_scale = scale

    @property
    def x_label(self):
        """The *x* axis label of the fit handled by this plot adapter. If ``'__del__'`` is used,
        the label will be set to :py:obj:`None`.

        :rtype: str
        """
        return self._axis_labels[0]

    @x_label.setter
    def x_label(self, label):
        if label == '__del__':
            label = None
        self._axis_labels[0] = label

    @property
    def y_label(self):
        """The *x* axis label of the fit handled by this plot adapter. If ``'__del__'`` is used,
        the label will be set to :py:obj:`None`.

        :rtype: str
        """
        return self._axis_labels[1]

    @y_label.setter
    def y_label(self, label):
        if label == '__del__':
            label = None
        self._axis_labels[1] = label

    @property
    def x_ticks(self):
        return self._ticks[0]

    @x_ticks.setter
    def x_ticks(self, ticks):
        self._ticks[0] = ticks

    @property
    def y_ticks(self):
        return self._ticks[1]

    @y_ticks.setter
    def y_ticks(self, ticks):
        self._ticks[1] = ticks

    @property
    def from_container(self):
        """Whether the contained fit object was created ad-hoc from just a data container."""
        return self._from_container

    @abc.abstractmethod
    def plot_data(self, target_axes, **kwargs):
        """Method called by the main plot routine to plot the data points to a specified
        :py:class:`matplotlib.axes.Axes` object.

        :param matplotlib.axes.Axes target_axes: The :py:mod:`matplotlib` target axes.
        :return: plot handle(s)
        """
        pass

    @abc.abstractmethod
    def plot_model(self, target_axes, **kwargs):
        """Method called by the main plot routine to plot the model to a specified
        :py:class:`matplotlib.axes.Axes` object.

        :param matplotlib.axes.Axes target_axes: The :py:mod:`matplotlib` target axes.
        :return: plot handle(s)
        """
        pass

    def plot_ratio(self, target_axes, error_contributions=('data',), **kwargs):
        """Plot the data/model ratio to a specified :py:obj:`matplotlib.axes.Axes` object.

        :param matplotlib.axes.Axes target_axes: The :py:obj:`matplotlib` axes used for plotting.
        :param error_contributions: Which error contributions to include when plotting the data.
            Can either be ``data``, ``'model'`` or both.
        :type error_contributions: str or Tuple[str]
        :param dict kwargs: Keyword arguments accepted by :py:obj:`matplotlib.pyplot.errorbar`.
        :return: plot handle(s)
        """

        _yerr = self._get_total_error(error_contributions)
        if _yerr is not None:
            _yerr /= self.model_y

        # TODO: how to handle case when x and y error/model differ?
        return target_axes.errorbar(
            self.data_x,
            self.data_y / self.model_y,
            xerr=self.data_xerr,
            yerr=_yerr,
            **kwargs
        )

    def plot_residual(self, target_axes, error_contributions=('data',), **kwargs):
        """Plot the residuals to a :py:obj:`matplotlib.axes.Axes` object.

        :param matplotlib.axes.Axes target_axes: The :py:obj:`matplotlib` axes used for plotting.
        :param error_contributions: Which error contributions to include when plotting the data.
            Can either be ``data``, ``'model'`` or both.
        :type error_contributions: str or Tuple[str]
        :param dict kwargs: Keyword arguments accepted by :py:obj:`matplotlib.pyplot.errorbar`.
        :return: plot handle(s)
        """
        # TODO: how to handle case when x and y error/model differ?
        return target_axes.errorbar(
            self.data_x,
            self.data_y - self.model_y,
            xerr=self.data_xerr,
            yerr=self._get_total_error(error_contributions),
            **kwargs
        )

    # Overridden by multi plot adapters
    def get_formatted_model_function(self, **kwargs):
        """return model function string"""
        return self._fit.model_function.formatter.get_formatted(**kwargs)

    # Overridden by multi plot adapters
    @property
    def model_function_parameter_formatters(self):
        """The model function parameter formatters, excluding the independent variable."""
        return self._fit.model_function.formatter.par_formatters

# -- must come last!


class Plot:
    """
    This is a class implementing the creation of Fits from one of more subclasses of
    :py:obj`PlotAdapterBase`. Consequently a :py:obj:`Plot` object manages one or several
    ``matplotlib`` figures. It controls the overall figure layout and is responsible for axes,
    subplot and legend management.
    """

    FIT_INFO_STRING_FORMAT_CHI2 = textwrap.dedent("""\
        {model_function}
        {parameters}
            $\\hookrightarrow${fit_quality}
            $\\hookrightarrow \\chi^2 \\, \\mathrm{{probability =}}${chi2_probability}
    """)
    FIT_INFO_STRING_FORMAT_SATURATED = textwrap.dedent("""\
        {model_function}
        {parameters}
            $\\hookrightarrow${fit_quality}
    """)
    FIT_INFO_STRING_FORMAT_NOT_SATURATED = textwrap.dedent("""\
        {model_function}
        {parameters}
            $\\hookrightarrow${cost}
            $\\hookrightarrow${fit_quality}
    """)

    def __init__(self, fit_objects, separate_figures=False):
        from kafe2.fit.tools.fit_wrapper import Fit

        # set the managed fit objects
        if isinstance(fit_objects, MultiFit):
            self._multifit = fit_objects
            fit_objects = fit_objects.fits
        else:
            self._multifit = None
        try:
            iter(fit_objects)
        except TypeError:
            fit_objects = (fit_objects,)
        self._from_container = tuple(isinstance(_fo, DataContainerBase) for _fo in fit_objects)
        self._fits = tuple(
            Fit(_fo) if _fc else _fo for _fc, _fo in zip(self._from_container, fit_objects))

        self._separate_figs = separate_figures

        # owned objects
        self._figure_dicts = []
        self._plot_adapters = None
        self._current_results = None

    # -- private methods

    def _get_axes(self, axes_key):
        try:
            return self._current_axes[axes_key]
        except KeyError:
            raise KeyError("No axes found for name '{}'!".format(axes_key))

    def _create_figure_axes(self, axes_keys, height_ratios=None, width_ratios=None, figsize=None):

        if height_ratios:
            assert len(axes_keys) == len(height_ratios)

        # plot axes layout
        _plot_axes_gs = gs.GridSpec(
            nrows=len(axes_keys),
            ncols=2,
            height_ratios=height_ratios,
            width_ratios=width_ratios,
        )

        # create figure
        self._current_figure = plt.figure(figsize=figsize)
        self._figure_dicts.append(dict(figure=self._current_figure))
        # 'tight_layout' has a bug in matplotlib < 2
        if not mpl.__version__.startswith('1'):
            self._current_figure.set_tight_layout(dict(h_pad=0.1))

        # create named axes
        self._current_axes = self._figure_dicts[-1]['axes'] = {
            _k : self._current_figure.add_subplot(_plot_axes_gs[_i, 0])
            for _i, _k in enumerate(axes_keys)
        }
        # create a fake axes for the legend
        self._current_axes['__legendfakeaxes__'] = self._current_figure.add_subplot(_plot_axes_gs[:, 1])
        self._current_axes['__legendfakeaxes__'].set_visible(False)

        # make all axes share the 'x' axis of the first axes
        for _ax_name, _ax in six.iteritems(self._current_axes):
            self._current_axes[axes_keys[0]].get_shared_x_axes().join(
                self._current_axes[axes_keys[0]], _ax)

        self._current_results = None  # populated on 'plot()'

    def _get_plot_adapters(self, plot_indices=None):
        '''retrieve plot adapters, creating them if needed'''

        plot_indices = plot_indices or range(len(self._fits))

        if self._plot_adapters is None:
            self._plot_adapters = []
            for _fit, _from_container in zip(self._fits, self._from_container):
                self._plot_adapters.append(
                    _fit.PLOT_ADAPTER_TYPE(_fit, _from_container)
                )
        if not self._separate_figs:
            _x_range_min = np.min([_pa.x_range[0] for _pa in self._plot_adapters])
            _x_range_max = np.max([_pa.x_range[1] for _pa in self._plot_adapters])
            for _plot_adapter in self._plot_adapters:
                _plot_adapter.x_range = (_x_range_min, _x_range_max)

        return [self._plot_adapters[_idx] for _idx in plot_indices]

    def _plot_and_get_results(self, plot_indices=None):
        plot_indices = plot_indices or range(len(self._fits))
        _plot_adapters = self._get_plot_adapters(plot_indices)
        if self._multifit is None:
            for _plot_adapter in _plot_adapters:
                if not _plot_adapter._fit.did_fit and not _plot_adapter.from_container:
                    warnings.warn(
                        "No fit has been performed for {}. Did you forget to run fit.do_fit()?"
                            .format(_plot_adapter._fit))
        elif not self._multifit.did_fit:
            warnings.warn(
                "No fit has been performed for {}. Did you forget to run fit.do_fit()?"
                    .format(self._multifit))

        _plots = {}
        for _i_pdc, _pdc in zip(plot_indices, _plot_adapters):

            if not _pdc.PLOT_SUBPLOT_TYPES:
                continue

            for _i_pt, (_pt, _pt_spec) in enumerate(six.iteritems(_pdc.PLOT_SUBPLOT_TYPES)):
                _axes_key = _pt_spec['target_axes']

                # skip plot elements meant for an inexistent axes
                if _axes_key not in self._current_axes:
                    continue

                _axes_plot_dicts = _plots.setdefault(_axes_key, {})

                _axes_plots = _axes_plot_dicts.setdefault('plots', [])

                _plot_kwargs = _pdc._get_subplot_kwargs(_i_pdc, _pt)
                if 'zorder' not in _plot_kwargs:
                    _plot_kwargs['zorder'] = 0
                _plot_kwargs['zorder'] = _plot_kwargs['zorder'] - 10*_i_pdc

                _artist = _pdc.call_plot_method(_pt,
                                                target_axes=self._get_axes(_axes_key),
                                                **_plot_kwargs
                                                )

                _axes_plots.append({
                    'type' : _pt,
                    'fit_index' : _i_pdc,
                    'adapter' : _pdc,
                    'artist' : _artist,
                })

                if _pdc.x_range is not None:
                    _xlim = _axes_plot_dicts.setdefault(
                        'x_range', _pdc.x_range)
                    _axes_plot_dicts['x_range'] = (
                        min(_xlim[0], _pdc.x_range[0]),
                        max(_xlim[1], _pdc.x_range[1])
                    )

                if _pdc.y_range is not None and _axes_key != 'ratio':  # y_range of ratio can be adjusted by plot kwargs
                    _ylim = _axes_plot_dicts.setdefault(
                        'y_range', _pdc.y_range)
                    _axes_plot_dicts['y_range'] = (
                        min(_ylim[0], _pdc.y_range[0]),
                        max(_ylim[1], _pdc.y_range[1])
                    )

        return _plots

    def _get_fit_info(self, plot_adapter, format_as_latex, asymmetric_parameter_errors):
        if self._multifit is None:
            plot_adapter._fit._update_parameter_formatters(
                update_asymmetric_errors=asymmetric_parameter_errors
            )
        else:
            if asymmetric_parameter_errors:
                self._multifit.asymmetric_parameter_errors
            self._multifit._update_parameter_formatters(
                update_asymmetric_errors=asymmetric_parameter_errors
            )

        _cost_func = plot_adapter._fit._cost_function  # TODO: public interface

        _ndf = plot_adapter._fit.ndf
        _cost_function_value = plot_adapter._fit.cost_function_value
        _gof_value = plot_adapter._fit.goodness_of_fit
        _info_format_dict = dict(
            model_function=plot_adapter.get_formatted_model_function(
                with_par_values=False,
                n_significant_digits=2,
                format_as_latex=format_as_latex,
                with_expression=True
            ),
            parameters='\n'.join([
                '    ' + _pf.get_formatted(
                    with_name=True,
                    with_value=True,
                    with_errors=plot_adapter._fit.errors_valid,
                    asymmetric_error=asymmetric_parameter_errors,
                    format_as_latex=format_as_latex
                )
                for _pf in plot_adapter.model_function_parameter_formatters
            ])
        )
        # _gof_value is None for UnbinnedFit and some user-defined cost functions.
        if plot_adapter._fit.errors_valid and _gof_value is not None:
            _info_format_dict["fit_quality"] = _cost_func.formatter.get_formatted(
                value=_gof_value,
                with_name=True,
                saturated=True,
                n_degrees_of_freedom=_ndf,
                with_value_per_ndf=True,
                format_as_latex=format_as_latex
            )
            if plot_adapter._fit._cost_function.is_chi2:
                _info_format_string = self.FIT_INFO_STRING_FORMAT_CHI2
                _chi2_pf = ParameterFormatter("chi2", plot_adapter._fit.chi2_probability)
                _info_format_dict["chi2_probability"] = _chi2_pf.get_formatted(
                    n_significant_digits=3, format_as_latex=format_as_latex)
            elif plot_adapter._fit._cost_function.saturated:
                _info_format_string = self.FIT_INFO_STRING_FORMAT_SATURATED
            else:
                _info_format_string = self.FIT_INFO_STRING_FORMAT_NOT_SATURATED
                _info_format_dict["cost"] = _cost_func.formatter.get_formatted(
                    value=_cost_function_value,
                    with_name=True,
                    format_as_latex=format_as_latex
                )
        else:
            _info_format_string = self.FIT_INFO_STRING_FORMAT_SATURATED
            _info_format_dict["fit_quality"] = _cost_func.formatter.get_formatted(
                value=_cost_function_value,
                with_name=True,
                format_as_latex=format_as_latex
            )

        _info_text = _info_format_string.format(**_info_format_dict)

        if self._multifit is not None:
            _multi_ndf = self._multifit.ndf
            _multi_cost_function = self._multifit._cost_function
            _multi_cost_function_value = self._multifit.cost_function_value
            _multi_gof = self._multifit.goodness_of_fit

            _multi_info_dict = dict()
            if _multi_cost_function.is_chi2:
                _template = "    $\\hookrightarrow$ global {fit_quality}\n"
                _chi2_pf = ParameterFormatter("chi2", self._multifit.chi2_probability)
                _multi_info_dict["chi2_probability"] = _chi2_pf.get_formatted(
                    n_significant_digits=3, format_as_latex=format_as_latex)
                _template += "    $\\hookrightarrow$ global $\\chi^2 \\, \\mathrm{{probability}} " \
                             "= ${chi2_probability}"
            elif _multi_cost_function.saturated:
                _template = "    $\\hookrightarrow$ global cost / ndf = {fit_quality}\n"
            else:
                _template = "    $\\hookrightarrow$ global cost = {cost}\n"
                _multi_info_dict["cost"] = _multi_cost_function.formatter.get_formatted(
                    value=_multi_cost_function_value,
                    with_name=False,
                    format_as_latex=format_as_latex
                )
                if _multi_gof is not None:
                    _template += "    $\\hookrightarrow$ global GoF / ndf = {fit_quality}\n"

            if _multi_gof is not None:
                _multi_info_dict["fit_quality"] = _multi_cost_function.formatter.get_formatted(
                    value=_multi_gof,
                    with_name=_multi_cost_function.is_chi2,
                    n_degrees_of_freedom=_multi_ndf,
                    with_value_per_ndf=True,
                    format_as_latex=format_as_latex
                )

            _info_text += _template.format(**_multi_info_dict)
        return _info_text

    def _render_legend(self, plot_results, axes_keys, fit_info=True, asymmetric_parameter_errors=False, **kwargs):
        """render the legend for axes `axes_keys`"""
        for _axes_key in axes_keys:
            _axes = self._get_axes(_axes_key)

            _hs_unsorted, _ls_unsorted = _axes.get_legend_handles_labels()
            _hs_sorted, _ls_sorted = [], []

            _axes_plots = plot_results[_axes_key]['plots']

            # -- go through each plot in order and generate the legend entry

            _prev_fit_index = None
            _fit_info = {}
            if isinstance(fit_info, bool):
                fit_info = [fit_info] * len(self._fits)
            for _i_plot, _plot_dict in enumerate(_axes_plots):
                # check if artist available for this plot
                try:
                    # if multiple artists were stored for a plot,
                    # only show the first  in the legend
                    try:
                        _artist_index = _hs_unsorted.index(_plot_dict['artist'][0])
                    except (ValueError, TypeError):
                        _artist_index = _hs_unsorted.index(_plot_dict['artist'])

                except (KeyError, ValueError):
                    # artist not available or not plottable -> skip
                    continue

                # append handle and label to the legend
                _hs_sorted.append(_hs_unsorted[_artist_index])
                _ls_sorted.append(_ls_unsorted[_artist_index])

                # if requested, compute info of fit associated to this artist
                _fit_index = _plot_dict['fit_index']
                _plot_adapter = _plot_dict['adapter']
                if fit_info[_fit_index] and not _plot_adapter.from_container:
                    if _fit_index not in _fit_info:
                        # compute fit info string (if not computed yet)
                        _fit_info[_fit_index] = dict(
                            text=self._get_fit_info(
                                _plot_adapter,
                                format_as_latex=True,
                                asymmetric_parameter_errors=asymmetric_parameter_errors,
                            )
                        )

                    # update the legend position at which to insert the text
                    _fit_info[_fit_index].update(
                        # put fit info directly after the last visible legend
                        # entry that corresponds to this fit
                        pos=len(_hs_sorted)
                    )

            # insert fit infos at the right positions
            for _i, _fi_dict in _fit_info.items():
                _hs_sorted.insert(_fi_dict['pos'] + _i, '_nokey_')
                _ls_sorted.insert(_fi_dict['pos'] + _i, _fi_dict['text'])

            # -- legend layout

            _zorder = kwargs.pop('zorder', 999)
            _bbox_to_anchor = kwargs.pop('bbox_to_anchor', None)
            if _bbox_to_anchor is None:
                _bbox_to_anchor = (1.05, 0.0, 0.67, 1.0)  # axes coordinates FIXME: no hardcoding!

            _mode = kwargs.pop('mode', "expand")
            _borderaxespad = kwargs.pop('borderaxespad', 0.1)
            _ncol = kwargs.pop('ncol', 1)

            kwargs['loc'] = 'upper left'

            # Note: legend must be attached to *figure*, not *axes*,
            # otherwise 'tight_layout' will consider it part of the axes
            # and produce undesirable layouts

            _leg = _axes.get_figure().legend(_hs_sorted, _ls_sorted,
                         mode=_mode,
                         borderaxespad=_borderaxespad,
                         ncol=_ncol,
                         handler_map={'_nokey_': DummyLegendHandler()},
                         **kwargs)
            _leg.set_zorder(_zorder)

            # manually change bbox from figure to axes
            _leg._bbox_to_anchor = self._get_axes('__legendfakeaxes__').bbox

    def _adjust_plot_ranges(self, plot_results):
        '''set the x and y ranges (all axes) to the total data range reported by the plot adapters'''
        for _axes_name, _axes_dict in six.iteritems(plot_results):
            _ax = self._get_axes(_axes_name)

            _xlim = _axes_dict.get('x_range', None)
            if _xlim:
                _ax.set_xlim(_xlim)

            _ylim = _axes_dict.get('y_range', None)
            if _ylim:
                _ax.set_ylim(_ylim)

    def _set_axis_labels(self, plot_results, axes_keys):
        """Set the x and y axis labels from given plot adapter labels"""
        for _axes_name, _axes_dict in six.iteritems(plot_results):
            _ax = self._get_axes(_axes_name)

            # collect different sets of axis labels
            _seen_x_labels = list()
            _seen_y_labels = list()
            for _plot in _axes_dict['plots']:
                _x_label = _plot['adapter'].x_label
                _y_label = _plot['adapter'].y_label
                if _x_label not in _seen_x_labels:
                    _seen_x_labels.append(_x_label)
                if _y_label not in _seen_y_labels:
                    _seen_y_labels.append(_y_label)

            # use concatenation of labels as axis label, filter to skip None labels
            _ax.set_xlabel(', '.join(filter(None, [label for label in _seen_x_labels])))
            _ax.set_ylabel(', '.join(filter(None, [label for label in _seen_y_labels])))

            _adapter = _axes_dict['plots'][0]['adapter']
            _x_ticks = _adapter.x_ticks
            _y_ticks = _adapter.y_ticks

            if len(_x_ticks) > 0:
                if isinstance(_x_ticks[0], str):
                    _ax.set_xticks(np.arange(len(_x_ticks)), labels=_x_ticks)
                else:
                    _ax.set_xticks(_x_ticks)

            if len(_y_ticks) > 0:
                _ax.set_yticks(_y_ticks)

        # hide x tick labels in all but the lowest axes
        for _key in axes_keys[:-1]:
            self._current_axes[_key].set_xlabel(None)
            for _label in self._current_axes[_key].get_xticklabels():
                _label.set_visible(False)

    # -- public properties

    @property
    def figures(self):
        """The ``matplotlib`` figures managed by this object."""
        return [_d['figure'] for _d in self._figure_dicts]

    @property
    def axes(self):
        """A list of dictionaries (one per figure) mapping names to
        ``matplotlib`` `Axes` objects contained in this figure."""
        return [_d['axes'] for _d in self._figure_dicts]

    @property
    def x_range(self):
        """The plotting x-range for each fit handled by this :py:obj:`~kafe2.Plot` object.
        :param: List of tuples containing the x_ranges for each fit.
        :type: list[tuple[float, float]] or tuple[float, float]
        """
        return [_adapter.x_range for _adapter in self._get_plot_adapters()]

    @x_range.setter
    def x_range(self, x_range):
        _adapters = self._get_plot_adapters()
        if np.ndim(x_range) == 1:
            if len(x_range) != 2:
                raise ValueError("x_range must contain two elements. A lower and an upper limit. "
                                 "Got {} elements".format(len(x_range)))
            x_range = itertools.repeat(x_range, len(_adapters))
        elif len(_adapters) != len(x_range):
            raise ValueError("Amount of x_ranges and fits does not match. Got {} x_ranges and have "
                             "{} fits".format(len(x_range), len(_adapters)))
        for i, _range in enumerate(x_range):
            _adapters[i].x_range = _range

    @property
    def y_range(self):
        """The plotting y-range for each fit handled by this :py:obj:`~kafe2.Plot` object.
        :param: List of tuples containing the y_ranges for each fit.
        :type: list[tuple[float, float]] or tuple[float, float]
        """
        return [_adapter.y_range for _adapter in self._get_plot_adapters()]

    @y_range.setter
    def y_range(self, y_range):
        _adapters = self._get_plot_adapters()
        if np.ndim(y_range) == 1:
            if len(y_range) != 2:
                raise ValueError("y_range must contain two elements. A lower and an upper limit. "
                                 "Got {} elements".format(len(y_range)))
            y_range = itertools.repeat(y_range, len(_adapters))
        elif len(_adapters) != len(y_range):
            raise ValueError("Amount of y_ranges and fits does not match. Got {} y_ranges and "
                             "have {} fits".format(len(y_range), len(_adapters)))
        for i, _range in enumerate(y_range):
            _adapters[i].y_range = _range

    def _repeat_str(self, var):
        """Repeat a string with the count of the given plot adapters. Used by setters."""
        _adapters = self._get_plot_adapters()
        if isinstance(var, str):  # check if string, if not probably list
            var = itertools.repeat(var, len(_adapters))
        elif len(_adapters) != len(var):
            raise ValueError("Length of input and fits does not match. Got {} inputs and have {} "
                             "fits".format(len(var), len(_adapters)))
        return var

    def _repeat_ticks(self, ticks):
        try:
            iter(ticks)
        except TypeError:
            raise TypeError(f"Ticks have to be iterable but received {ticks}")
        try:
            iter(ticks[0])
            _ticks_is_1d_iter = isinstance(ticks[0], str)
        except TypeError:
            _ticks_is_1d_iter = True
        if _ticks_is_1d_iter:
            ticks = itertools.repeat(ticks, len(self._get_plot_adapters()))
        return ticks

    @property
    def x_scale(self):
        """The x-scale for each fit used for creating the support values when plotting and axis
        scaling.
        :type: list[str] or str
        """
        return [_adapter.x_scale for _adapter in self._get_plot_adapters()]

    @x_scale.setter
    def x_scale(self, x_scale):
        x_scale = self._repeat_str(x_scale)
        for adapter, _scale in zip(self._get_plot_adapters(), x_scale):
            adapter.x_scale = _scale

    @property
    def y_scale(self):
        """The y-scale for each fit used when plotting.
        :type: list[str] or str
        """
        return [_adapter.y_scale for _adapter in self._get_plot_adapters()]

    @y_scale.setter
    def y_scale(self, y_scale):
        y_scale = self._repeat_str(y_scale)
        for adapter, _scale in zip(self._get_plot_adapters(), y_scale):
            adapter.y_scale = _scale

    @property
    def x_label(self):
        """The x-label(s) of the plot. If multiple fits are handled by this plot this is a list of
        strings. Multiple labels will be separated by a comma in the final plot while skipping
        duplicates.
        If a label is :py:obJ:`None` or ``'__del__'`` it will be removed.
        :type: str or list[str]"""
        return [_adapter.x_label for _adapter in self._get_plot_adapters()]

    @x_label.setter
    def x_label(self, label):
        label = self._repeat_str(label)
        for _l, adapter in zip(label, self._get_plot_adapters()):
            adapter.x_label = _l

    @property
    def y_label(self):
        """The y-label(s) of the plot. If multiple fits are handled by this plot this is a list of
        strings. Multiple labels will be separated by a comma in the final plot while skipping
        duplicates.
        If a label is :py:obJ:`None` or ``'__del__'`` it will be removed.
        :type: str or list[str]"""
        return [_adapter.y_label for _adapter in self._get_plot_adapters()]

    @y_label.setter
    def y_label(self, label):
        label = self._repeat_str(label)
        for _l, adapter in zip(label, self._get_plot_adapters()):
            adapter.y_label = _l

    @property
    def x_ticks(self):
        return [_adapter.x_ticks for _adapter in self._get_plot_adapters()]

    @x_ticks.setter
    def x_ticks(self, ticks):
        for _t, adapter in zip(self._repeat_ticks(ticks), self._get_plot_adapters()):
            adapter.x_ticks = _t

    @property
    def y_ticks(self):
        return [_adapter.y_ticks for _adapter in self._get_plot_adapters()]

    @y_ticks.setter
    def y_ticks(self, ticks):
        for _t, adapter in zip(self._repeat_ticks(ticks), self._get_plot_adapters()):
            adapter.y_ticks = _t

    # public methods

    @staticmethod
    def show(*args, **kwargs):
        """Convenience wrapper for matplotlib.pyplot.show()"""
        plt.show(*args, **kwargs)


    def plot(self, legend=True, fit_info=True, asymmetric_parameter_errors=False,
             ratio=False, ratio_range=None, ratio_height_share=0.25,
             residual=False, residual_range=None, residual_height_share=0.25,
             plot_width_share=0.5, figsize=None):
        """
        Plot data, model (and other subplots) for all child :py:obj:`Fit` objects.

        :param legend: if ``True``, a legend is rendered
        :param fit_info: If :py:obj:`True`, fit results will be shown in the legend. This can also
            be a list of booleans, corresponding to the fits handled by this
            :py:obj:`~.Plot`-object.
        :type legend: bool or typing.Collection[bool]
        :param asymmetric_parameter_errors: if ``True``, parameter errors in fit results will be asymmetric
        :param ratio: if ``True``, a secondary plot containing data/model ratios is shown below the main plot
        :param ratio_range: the *y* range to set in the secondary plot
        :type ratio_range: tuple of 2 floats
        :param ratio_height_share: share of the total height to be taken up by the secondary plot
        :type ratio_height_share: float
        :param plot_width_share: share of the total width to be taken up by the plot(s)
        :type plot_width_share: float
        :param figsize: the (*width*, *height*) of the figure (in inches) or ``None`` to use default
        :type figsize: tuple of 2 floats

        :return: dictionary containing information about the plotted objects
        :rtype: dict
        """
        if ratio and residual:
            raise NotImplementedError("Cannot plot ratio and residual at the same time.")

        with rc_context(kafe2_rc):
            _axes_keys = ('main',)
            _height_ratios = [1.0]
            _width_ratios = (plot_width_share, 1.0 - plot_width_share)

            if ratio:
                _axes_keys += ('ratio',)
                _height_ratios[0] -= ratio_height_share
                _height_ratios.append(ratio_height_share)
            elif residual:
                _axes_keys += ('residual',)
                _height_ratios[0] -= residual_height_share
                _height_ratios.append(residual_height_share)

            _all_plot_results = []
            for i in range(len(self._fits) if self._separate_figs else 1):
                self._create_figure_axes(
                    _axes_keys,
                    width_ratios=_width_ratios,
                    height_ratios=_height_ratios,
                    figsize=figsize,
                )

                _plot_results = self._plot_and_get_results(plot_indices=(i,) if self._separate_figs else None)

                # set axis scales for the main plot, x-axis is shared with ratio plot
                self._get_axes('main').set_xscale(self.x_scale[i])
                self._get_axes('main').set_yscale(self.y_scale[i])

                if legend:
                    self._render_legend(plot_results=_plot_results, axes_keys=('main',), fit_info=fit_info,
                                        asymmetric_parameter_errors=asymmetric_parameter_errors)

                self._adjust_plot_ranges(_plot_results)
                self._set_axis_labels(_plot_results, axes_keys=_axes_keys)
                try:
                    self._current_figure.align_ylabels()
                except AttributeError:
                    # matplotlib < 2.0.0
                    pass

                if ratio:
                    _axis = self._current_axes['ratio']
                    _ratio_label = kc('fit', 'plot', 'ratio_label')
                    _axis.set_ylabel(_ratio_label)
                    if ratio_range is None:
                        _plot_adapters = (self._get_plot_adapters()[i:i+1] if self._separate_figs
                                          else self._get_plot_adapters())
                        _max_abs_deviation = 0
                        for _plot_adapter in _plot_adapters:
                            _max_abs_deviation = max(_max_abs_deviation, np.max(
                                (
                                    np.abs(_plot_adapter.data_yerr)
                                    + np.abs(_plot_adapter.data_y - _plot_adapter.model_y)
                                ) / np.abs(_plot_adapter.model_y)
                            ))
                        # Small gap between highest error bar and plot border:
                        _low = 1 - _max_abs_deviation * 1.05
                        _high = 1 + _max_abs_deviation * 1.05
                        _axis.set_ylim((_low, _high))
                    else:
                        _axis.set_ylim(ratio_range)
                if residual:
                    _axis = self._current_axes['residual']
                    _residual_label = kc('fit', 'plot', 'residual_label')
                    _axis.set_ylabel(_residual_label)
                    if residual_range is None:
                        _plot_adapters = (self._get_plot_adapters()[i:i+1] if self._separate_figs
                                          else self._get_plot_adapters())
                        _max_abs_deviation = 0
                        for _plot_adapter in _plot_adapters:
                            _max_abs_deviation = max(_max_abs_deviation, np.max(
                                np.abs(_plot_adapter.data_yerr)
                                + np.abs(_plot_adapter.data_y - _plot_adapter.model_y)
                            ))
                        # Small gap between highest error bar and plot border:
                        _low = -_max_abs_deviation * 1.05
                        _high = _max_abs_deviation * 1.05
                        _axis.set_ylim((_low, _high))
                    else:
                        _axis.set_ylim(residual_range)

                _all_plot_results.append(_plot_results)

            self._current_results = _all_plot_results

            return _all_plot_results

    def get_keywords(self, plot_type):
        """Retrieve keyword arguments for plots with type `plot_type` as they would be used when calling `plot`.

        This is an advanced function. An understanding of how plotting with
        `matplotlib` and the `PlotAdapter` classes in *kafe2* work is recommended.

        The `plot_type` must be one of the plot types registered in the
        `PlotAdapter` (e.g. ``'data'``, ``'model_line'`` etc.).

        :param plot_type: keyword identifying the plot type for which to set a custom keyword argument
        :type plot_type: str

        :return: list of dictionaries (one per fit instance) containing plot keywords and their values
        :rtype: list of dict
        """

        _adapters = self._get_plot_adapters()

        _keywords_list = []
        for _i_pdc, _pdc in enumerate(_adapters):
            _keywords_list.append(
                _pdc._get_subplot_kwargs(
                    _i_pdc,
                    plot_type
                ))

        return _keywords_list

    def set_keywords(self, plot_type, keyword_spec):
        """Set values for keyword arguments used for plots with type `plot_type`.

        This is an advanced function. An understanding of how plotting with
        `matplotlib` and the `PlotAdapter` classes in *kafe2* work is recommended.

        The `plot_type` must be one of the plot types registered in the
        `PlotAdapter` (e.g. ``'data'``, ``'model_line'`` etc.).

        The `keyword_spec` contains dictionaries whose contents will be passed as
        keyword arguments to the plot adapter method responsible for plotting the
        `plot_type`. If `keyword` spec contains a key for which a default value is
        configured, it will be overridden.

        Passing the following special values for a keyword will have the following effects:

        * ``'__del__'``: the value will be removed from the keyword arguments. This includes
          default values, meaning that the plot call will be made **without** the keyword
          argument even if a default value for it exists.
        * ``'__default__'``: the customized value will be replaced by the default value.

        .. note::

            No keyword/value validation is done: everything is passed to the underlying plot
            methods as specified. Incorrect or incompatible keywords may be ignored or lead to errors.

        As an example, to override the labels shown in the legend entries for the `data`

        .. code:: python

            p = Plot([fit_1, fit_2])
            p.customize('data', [dict(label='My Data Label'), dict(label='Another Data Label')])

        To set keywords for a single `fit`, pass values as ``(index, value)``, where `index` is
        the index of the `fit` object:

        .. code:: python

            p = Plot([fit_1, fit_2])
            p.customize('data', [(1, dict(label='Another Data Label'))])

        :param plot_type: keyword identifying the plot type for which to set a custom keyword argument
        :type plot_type: str
        :param keyword_spec: specification of dictionaries containing the keyword arguments to use for fit.
            Can be either a list of dictionaries with a length corresponding to the number of `fit` objects
            managed by this `Plot` instance, or a list of tuples of the form ``(index, dict)``, where
            ``index``  denotes the index of the `fit` object for which the dictionary `dict` should be used.
        :type keyword_spec: list of values or list of 2-tuples like ``(index, value)``

        :return: this `Plot` instance
        :rtype: `Plot`
        """

        _adapters = self._get_plot_adapters()

        _has_tuples = None
        for _spec in keyword_spec:
            if isinstance(_spec, tuple):
                # raise if both tuples and non-tuples provided
                if _has_tuples is None:
                    _has_tuples = True
                elif not _has_tuples:
                    raise ValueError(
                        "Cannot set custom plot keyword arguments: "
                        "provided `values` contain a mix of tuples and non-tuples!")

                # validate tuple
                if not len(_spec) == 2:
                    raise ValueError(
                        "Cannot set custom plot keyword arguments: "
                        "tuple {!r} has length {} (expected "
                        "2) ".format(_spec, len(_spec)))

                # validate index
                try:
                    _adapters[_spec[0]]
                except IndexError:
                    raise ValueError(
                        "Cannot set custom plot keyword arguments: "
                        "invalid index {!r} encountered (object manages {} fit "
                        "objects)!".format(_spec[0], len(_adapters)))
            else:
                if _has_tuples is None:
                    _has_tuples = False
                elif _has_tuples:
                    raise ValueError(
                        "Cannot set custom plot keyword arguments: "
                        "provided `values` contain a mix of tuples and non-tuples!")

        if not _has_tuples:
            # must provide same amount of values as fits
            if len(_adapters) != len(keyword_spec):
                raise ValueError(
                    "Cannot set custom plot keyword argument: "
                    "{} values provided but this object manages {} "
                    "fit objects!".format(len(keyword_spec), len(_adapters)))

            # convert plain list to tuple
            keyword_spec = [(_index, _dict) for _index, _dict in enumerate(keyword_spec)]
        if _has_tuples:
            # can provide less tuples than fits but not more
            if len(_adapters) < len(keyword_spec):
                raise ValueError(
                    "Cannot set custom plot keyword argument: "
                    "{} values provided but this object manages {} "
                    "fit objects!".format(len(keyword_spec), len(_adapters)))

        for _index, _dict in keyword_spec:
            _adapters[_index].update_plot_kwargs(plot_type, _dict)

        return self  # allow chaining calls

    def customize(self, plot_type, keyword, values):
        """Set values for keyword arguments used for plots with type `plot_type`.

        This is a convenience wrapper around `set_keywords`.

        The `keyword` will be passed to the plot adapter method responsible for
        plotting the `plot_type` as a keyword argument with a value taken
        from `values`. If a default value for `keyword` is configured, it is
        overridden.

        The `values` can be specified in two ways:

        #. as a list with a length corresponding to the number of `fit` objects managed
           by this `Plot` instance. The special value ``'__skip__'`` can be used to skip
           `fit` objects.
        #. as a list of tuples of the form ``(index, value)``, where `index` denotes the
           index of the `fit` object for which the `value` should be used.

        Passing the following special values for a keyword will have the following effects:

        * ``'__del__'``: the value will be removed from the keyword arguments. This includes
          default values, meaning that the plot call will be made **without** the keyword
          argument even if a default value for it exists.
        * ``'__default__'``: the customized value will be replaced by the default value.
        * ``'__skip__'``: the keywords for this `fit` will not be changed.

        .. note::

            No keyword/value validation is done: everything is passed to the underlying plot
            methods as specified. Incorrect or incompatible keywords may be ignored or lead to errors.

        As an example, to override the labels shown in the legend entries for the `data`

        .. code:: python

            p = Plot([fit_1, fit_2])
            p.customize('data', 'label', ['My Data Label', 'Another Data Label'])

        To set keywords for a single `fit`, pass values as ``(index, value)``, where `index` is
        the index of the `fit` object:

        .. code:: python

            p = Plot([fit_1, fit_2])
            p.customize('data', 'label', [(1, 'Another Data Label')])

        :param plot_type: keyword identifying the plot type for which to set a custom keyword argument
        :type plot_type: str
        :param keyword: the keyword argument. The corresponding value in `values` will be passed
            to the plot adapter method using this keyword argument
        :type keyword: str
        :param values: values that the keyword argument should take for each fit. Can be a list of values
            with a length corresponding to the number of `fit` objects managed by this `Plot` instance,
            or a list of tuples of the form ``(index, value)``
        :type values: list of values or list of 2-tuples like ``(index, value)``

        :return: this `Plot` instance
        :rtype: `Plot`
        """
        if values is None or isinstance(values, (int, float, str)):
            values = (values,) * len(self._fits)

        _has_tuples = None
        for _val in values:
            if isinstance(_val, tuple):
                # raise if both tuples and non-tuples provided
                if _has_tuples is None:
                    _has_tuples = True
                elif not _has_tuples:
                    raise ValueError(
                        "Cannot set custom plot keyword arguments: "
                        "provided `values` contain a mix of tuples and non-tuples!")

                # validate tuple
                if not len(_val) == 2:
                    raise ValueError(
                        "Cannot set custom plot keyword arguments: "
                        "tuple {!r} has length {} (expected "
                        "2) ".format(_val, len(_val)))
            else:
                if _has_tuples is None:
                    _has_tuples = False
                elif _has_tuples:
                    raise ValueError(
                        "Cannot set custom plot keyword arguments: "
                        "provided `values` contain a mix of tuples and non-tuples!")

        if not _has_tuples:
            _dicts = [
                {keyword: _value} if _value != '__skip__' else {}
                for _value in values
            ]
        else:
            _dicts = [
                (_index, {keyword: _value})
                for _index, _value in values
                if _value != '__skip__'
            ]

        return self.set_keywords(plot_type, _dicts)

    def save(self, fname=None, figures='all', *args, **kwargs):
        """
        Saves the plot figures to files.
        args and kwargs are passed on to matplotlib.Figure.savefig() .

        :param fname: Output file name(s), defaults to fit.png or fit_0.png, fit_1.png, ...
        :type fname: None or str or iterable of str.
        :param figures: Which figures to save.
        :type figures: 'all' or int or iterable of int
        """
        if figures == 'all':
            _figure_indices = range(len(self._figure_dicts))
        elif isinstance(figures, int):
            _figure_indices = [figures]
        _file_name_list = fname
        if fname is None:
            if len(self._figure_dicts) == 1:
                _file_name_list = ["fit.png"]
            else:
                _file_name_list = [f'fit_{_i}.png' for _i in _figure_indices]
        elif isinstance(fname, str):
            if len(_figure_indices) == 1:
                _file_name_list = [fname]
            else:
                name, extension = os.path.splitext(fname)
                _file_name_list = [f'{name}_{_i}{extension}' for _i in _figure_indices]
        elif isinstance(fname, Iterable):
            _file_name_list = fname
        else:
            raise ValueError(
                "fname must be either None, a string, or an iterable, but received type "
                f"{type(fname)}"
            )
        if len(_file_name_list) != len(_figure_indices):
            raise ValueError(
                f'Received {len(_file_name_list)} file names for {len(_figure_indices)} figures.'
                'Must be the same!'
            )
        for _figure_index, _file_name in zip(_figure_indices, _file_name_list):
            self._figure_dicts[_figure_index]['figure'].savefig(_file_name, *args, **kwargs)



import numpy as np

from .._base import PlotAdapterBase
from .._aux import step_fill_between


__all__ = ["IndexedPlotAdapter"]


class IndexedPlotAdapter(PlotAdapterBase):

    PLOT_STYLE_CONFIG_DATA_TYPE = 'indexed'

    PLOT_SUBPLOT_TYPES = dict(
        PlotAdapterBase.PLOT_SUBPLOT_TYPES,
    )

    def __init__(self, indexed_fit_object, from_container=False):
        """
        Construct an :py:obj:`IndexedPlotContainer` for a :py:obj:`~kafe2.fit.indexed.IndexedFit` object:

        :param fit_object: an :py:obj:`~kafe2.fit.indexed.IndexedFit` object
        :param from_container: Whether indexed_fit_object was created ad-hoc from just a data
            container.
        :type from_container: bool
        """
        super(IndexedPlotAdapter, self).__init__(
            fit_object=indexed_fit_object, from_container=from_container)
        self.x_range = (-0.5, self._fit.data_size - 0.5)

    @property
    def data_x(self):
        """data x values"""
        return np.arange(self._fit.data_size)

    @property
    def data_y(self):
        """data y values"""
        return self._fit.data

    @property
    def data_xerr(self):
        """x error bars for data: ``None`` for :py:obj:`IndexedPlotContainer`"""
        return None

    @property
    def data_yerr(self):
        """y error bars for data: total uncertainty"""
        return self._fit.total_error

    @property
    def model_x(self):
        """model prediction x values"""
        return self.data_x

    @property
    def model_y(self):
        """model prediction y values"""
        return self._fit.model

    @property
    def model_xerr(self):
        """x error bars for model (actually used to represent the horizontal step length)"""
        return 0.5

    @property
    def model_yerr(self):
        """y error bars for model: ``None`` for :py:obj:`IndexedPlotContainer`"""
        return None #self.fit.model_error

    # public methods

    def plot_data(self, target_axes, **kwargs):
        """
        Plot the measurement data to a specified ``matplotlib`` ``Axes`` object.

        :param target_axes: ``matplotlib`` ``Axes`` object
        :param kwargs: keyword arguments accepted by the ``matplotlib`` methods ``errorbar`` or ``plot``
        :return: plot handle(s)
        """
        if self._fit.has_errors:
            _yerr = np.sqrt(self.data_yerr ** 2
                            + self._fit._cost_function.get_uncertainty_gaussian_approximation(self.data_y) ** 2)
            return target_axes.errorbar(self.data_x, self.data_y, xerr=self.data_xerr, yerr=_yerr, **kwargs)
        _yerr = self._fit._cost_function.get_uncertainty_gaussian_approximation(self.data_y)
        if np.all(_yerr == 0):
            return target_axes.plot(self.data_x, self.data_y, **kwargs)
        return target_axes.errorbar(self.data_x, self.data_y, yerr=_yerr, **kwargs)

    def plot_model(self, target_axes, **kwargs):
        """
        Plot the model predictions to a specified matplotlib ``Axes`` object.

        :param target_axes: ``matplotlib`` ``Axes`` object
        :param kwargs: keyword arguments accepted by the :py:func:`~kafe2._aux.step_fill_between` method
        :return: plot handle(s)
        """
        return step_fill_between(target_axes,
                                 self.model_x,
                                 self.model_y,
                                 xerr=self.model_xerr,
                                 yerr=self.model_yerr,
                                 draw_central_value=True,
                                 continuous=False,
                                 **kwargs
                                 )




from __future__ import print_function

import logging

from .minimizer_base import MinimizerBase
from ..contour import ContourFactory

try:
    import scipy.optimize as opt
except ImportError:
    # TODO: handle importing nonexistent minimizer
    raise

import numpy as np
import numdifftools as nd


class MinimizerScipyOptimize(MinimizerBase):
    def __init__(self,
                 parameter_names, parameter_values, parameter_errors,
                 function_to_minimize, tolerance=1e-6, errordef=MinimizerBase.ERRORDEF_CHI2,
                 method=None):
        self._method = method
        self._par_bounds = None
        self._par_fixed = np.array([False] * len(parameter_names))
        self._par_constraints = []

        self._opt_result = None
        self._x0 = None  # Stores initial value for x0 when profiling a parameter
        super(MinimizerScipyOptimize, self).__init__(
            parameter_names=parameter_names, parameter_values=parameter_values,
            parameter_errors=parameter_errors, function_to_minimize=function_to_minimize,
            tolerance=tolerance, errordef=errordef
        )

    # -- private methods

    def _save_state(self):
        if self._par_val is None:
            self._save_state_dict['parameter_values'] = self._par_val
        else:
            self._save_state_dict['parameter_values'] = np.array(self._par_val)
        if self._par_err is None:
            self._save_state_dict['parameter_errors'] = self._par_err
        else:
            self._save_state_dict['parameter_errors'] = np.array(self._par_err)
        if self._par_bounds is None:
            self._save_state_dict['parameter_bounds'] = self._par_bounds
        else:
            self._save_state_dict['parameter_bounds'] = np.array(self._par_bounds)
        self._save_state_dict['function_value'] = self._fval
        self._save_state_dict['par_fixed'] = np.array(self._par_fixed)
        self._save_state_dict['opt_result'] = self._opt_result
        super(MinimizerScipyOptimize, self)._save_state()

    def _load_state(self):
        self._par_val = self._save_state_dict['parameter_values']
        if self._par_val is not None:
            self._par_val = np.array(self._par_val)
        self._par_err = self._save_state_dict['parameter_errors']
        if self._par_err is not None:
            self._par_err = np.array(self._par_err)
        self._par_bounds = self._save_state_dict['parameter_bounds']
        if self._par_bounds is not None:
            self._par_bounds = np.array(self._par_bounds)
        self._fval = self._save_state_dict['function_value']
        self._par_fixed = np.array(self._save_state_dict['par_fixed'])
        self._opt_result = self._save_state_dict['opt_result']
        super(MinimizerScipyOptimize, self)._load_state()

    # -- public properties

    @property
    def parameter_values(self):
        return np.array(self._par_val)

    @parameter_values.setter
    def parameter_values(self, new_values):
        self._par_val = np.array(new_values)
        self.reset()

    @property
    def parameter_errors(self):
        return self._par_err.copy()

    @parameter_errors.setter
    def parameter_errors(self, new_errors):
        _err_array = np.array(new_errors)
        if not np.all(_err_array > 0):
            raise ValueError("All parameter errors must be > 0! Received: %s" % new_errors)
        self._par_err = _err_array
        self.reset()

    # -- private "properties"

    # -- public methods

    def set(self, parameter_name, parameter_value):
        if parameter_name not in self._par_names:
            raise ValueError("No parameter named '%s'!" % (parameter_name,))
        _par_id = self._par_names.index(parameter_name)
        self._par_val[_par_id] = parameter_value
        self.reset()

    def fix(self, parameter_name):
        _par_id = self._par_names.index(parameter_name)
        self._par_fixed[_par_id] = True
        self._invalidate_cache()

    def is_fixed(self, parameter_name):
        _par_id = self._par_names.index(parameter_name)
        return self._par_fixed[_par_id]

    def release(self, parameter_name):
        _par_id = self._par_names.index(parameter_name)
        self._par_fixed[_par_id] = False
        self._invalidate_cache()

    def limit(self, parameter_name, parameter_bounds):
        assert len(parameter_bounds) == 2
        _par_id = self._par_names.index(parameter_name)
        if parameter_bounds[0] is not None and self._par_val[_par_id] < parameter_bounds[0]:
            self.set(parameter_name, parameter_bounds[0])
        elif parameter_bounds[1] is not None and self._par_val[_par_id] > parameter_bounds[1]:
            self.set(parameter_name, parameter_bounds[1])
        if self._par_bounds is None:
            self._par_bounds = [(None, None) for _pn in self._par_names]
        self._par_bounds[_par_id] = parameter_bounds

    def unlimit(self, parameter_name):
        _par_id = self._par_names.index(parameter_name)
        self._par_bounds[_par_id] = (None, None)
        _all_pars_unbounded = True
        for _par_bound in self._par_bounds:
            if _par_bound != (None, None):
                _all_pars_unbounded = False
                break
        if _all_pars_unbounded:
            self._par_bounds = None

    def minimize(self, max_calls=6000):
        if np.all(self._par_fixed):
            raise RuntimeError("Cannot perform a fit if all parameters are fixed!")
        if np.any(self._par_fixed):
            # if pars are fixed arg list becomes shorter -> pick and insert fixed pars from self.parameter_values
            _par_fixed_indices = np.array(self._par_fixed, dtype=int)  # 1 if fixed, 0 otherwise
            # keep track of the arg positions within the fixed/dynamic lists:
            _position_indices = np.zeros_like(self.parameter_values, dtype=int)
            _n_fixed_parameters = 0
            _par_vals = []
            for _par_index, _par_fixed in enumerate(self._par_fixed):
                if _par_fixed:
                    _position_indices[_par_index] = _par_index
                    _n_fixed_parameters += 1
                else:
                    _position_indices[_par_index] = _par_index - _n_fixed_parameters
                    _par_vals.append(self.parameter_values[_par_index])
            _dyn_and_fixed_args = np.zeros(shape=(2,) + self.parameter_values.shape)
            _dyn_and_fixed_args[1] = self.parameter_values

            def _func(args):
                _dyn_and_fixed_args[0, 0:-_n_fixed_parameters] = args
                _selected_values = _dyn_and_fixed_args[_par_fixed_indices, _position_indices]
                return self._func_wrapper_unpack_args(_selected_values)

            if self._par_bounds is None:
                _par_bounds = None
            else:
                _par_bounds = []
                for _par_index, _par_fixed in enumerate(self._par_fixed):
                    if not _par_fixed:
                        _par_bounds.append(self._par_bounds[_par_index])

        else:
            _func = self._func_wrapper_unpack_args
            _par_vals = self.parameter_values
            _par_bounds = self._par_bounds

        disp = False
        if logging.root.level <= logging.INFO:
            disp = True

        self._opt_result = opt.minimize(_func,
                                        _par_vals,
                                        args=(),
                                        method=self._method,
                                        jac=None,
                                        hess=None, hessp=None,
                                        bounds=_par_bounds,
                                        constraints=self._par_constraints,
                                        tol=self.tolerance,
                                        callback=None,
                                        options=dict(maxiter=max_calls, disp=disp))
        self._did_fit = True
        self._invalidate_cache()

        if np.any(self._par_fixed):
            _dyn_and_fixed_args[0, 0:-_n_fixed_parameters] = self._opt_result.x
            self._par_val = _dyn_and_fixed_args[_par_fixed_indices, _position_indices]
        else:
            self._par_val = self._opt_result.x

        self._fval = self._opt_result.fun

        # Write back parameter values to nexus parameter nodes:
        self._func_wrapper_unpack_args(self.parameter_values)

        # Update parameter errors.
        # This is not done lazily because parameter errors need to be persistent.
        self._par_err = np.sqrt(np.diag(self.cov_mat))

    def contour(self, parameter_name_1, parameter_name_2, sigma=1.0, **minimizer_contour_kwargs):
        if not self.did_fit:
            raise RuntimeError("Need to perform a fit before calling contour()!")
        _algorithm = minimizer_contour_kwargs.pop("algorithm", "heuristic_grid")

        if _algorithm == "beacon":
            pass
        elif _algorithm == "heuristic_grid":
            _initial_points = minimizer_contour_kwargs.pop("initial_points", 1)
            _iterations = minimizer_contour_kwargs.pop("iterations", 5)
            _area_scale_factor = minimizer_contour_kwargs.pop("area_scale_factor", 1.5)
        else:
            raise ValueError("Unknown algorithm: {}".format(_algorithm))

        if minimizer_contour_kwargs:
            raise ValueError(
                "Unknown parameters for {}: {}".format(_algorithm, minimizer_contour_kwargs.keys()))

        if _algorithm == "beacon":
            return self._contour_beacon(parameter_name_1, parameter_name_2, sigma=sigma)
        if _algorithm == "heuristic_grid":
            return self._contour_heuristic_grid(parameter_name_1, parameter_name_2, sigma=sigma,
                                                initial_points=_initial_points, iterations=_iterations,
                                                area_scale_factor=_area_scale_factor)

    def _contour_heuristic_grid(self, parameter_name_1, parameter_name_2, sigma=1.0, initial_points=1,
                                iterations=5, area_scale_factor=1.5):
        initial_points = int(initial_points)
        iterations = int(iterations)

        if initial_points < 1:
            raise MinimizerScipyOptimize("initial_points must be a >= 1")
        if iterations < 0:
            raise MinimizerScipyOptimize("iterations must be a >= 0")

        _initial_points_per_axis = 1 + initial_points * 2
        _target_points_per_axis = 1 + initial_points * 2 ** (iterations + 1)
        _ids = (self._par_names.index(parameter_name_1), self._par_names.index(parameter_name_2))
        _minimum = np.asarray([self._par_val[_ids[0]], self._par_val[_ids[1]]])
        _err = np.asarray([self.parameter_errors[_ids[0]], self.parameter_errors[_ids[1]]])

        _x_values = np.linspace(start=-area_scale_factor * sigma * _err[0], stop=area_scale_factor * sigma * _err[0],
                                num=_target_points_per_axis, endpoint=True)
        _x_values += _minimum[0]
        _y_values = np.linspace(start=-area_scale_factor * sigma * _err[1], stop=area_scale_factor * sigma * _err[1],
                                num=_target_points_per_axis, endpoint=True)
        _y_values += _minimum[1]

        _grid = np.zeros((_target_points_per_axis, _target_points_per_axis)) - 1
        _x_step = int((_target_points_per_axis - 1) / (_initial_points_per_axis - 1))
        _y_step = int((_target_points_per_axis - 1) / (_initial_points_per_axis - 1))

        _min_coords = int((_target_points_per_axis - 1) / 2)
        _confirmed_coords = set()
        _unsure_coords = set()

        for _x in range(0, _target_points_per_axis, _x_step):
            for _y in range(0, _target_points_per_axis, _y_step):
                _local_constraints = [{'type': 'eq', 'fun': lambda x: x[_ids[0]] - _x_values[_x]},
                                      {'type': 'eq', 'fun': lambda x: x[_ids[1]] - _y_values[_y]}]
                _grid[_x, _y] = self._calc_fun_with_constraints(_local_constraints)

        _min_fun = min(self.function_value, _grid[_min_coords, _min_coords])
        _contour_fun = _min_fun + sigma ** 2

        _iterations = 0
        while _x_step > 0 and _y_step > 1:
            if _iterations % 2 == 0:
                _x_0 = int(_x_step / 2)
                _y_0 = int(_y_step / 2)
                _vector_1 = (int(_x_step / 2), int(_y_step / 2))
                _vector_2 = (int(_x_step / 2), -int(_y_step / 2))
            else:
                _x_0 = 0
                _y_0 = 0
                _vector_1 = (_x_step, 0)
                _vector_2 = (0, int(_y_step / 2))

            for _x in range(_x_0, _target_points_per_axis, _x_step):
                if _iterations % 2 == 1 and _x % (2 * _x_step) == 0:
                    _current_y_0 = _y_0 + int(_y_step / 2)
                else:
                    _current_y_0 = _y_0
                for _y in range(_current_y_0, _target_points_per_axis, _y_step):
                    _point_value = self._heuristic_point_evaluation(_contour_fun, _grid, _x, _y, _vector_1, _vector_2)
                    if _point_value == -1:
                        _local_constraints = [{'type': 'eq', 'fun': lambda x: x[_ids[0]] - _x_values[_x]},
                                              {'type': 'eq', 'fun': lambda x: x[_ids[1]] - _y_values[_y]}]
                        _grid[_x, _y] = self._calc_fun_with_constraints(_local_constraints)
                        _confirmed_coords.add((_x, _y))
                        if _iterations % 2 == 0:
                            _unsure_coords.add((_x - _x_step, _y))
                            _unsure_coords.add((_x, _y - _y_step))
                            _unsure_coords.add((_x + _x_step, _y))
                            _unsure_coords.add((_x, _y + _y_step))
                        else:
                            _unsure_coords.add((_x - _x_step, _y - int(_y_step / 2)))
                            _unsure_coords.add((_x - _x_step, _y + int(_y_step / 2)))
                            _unsure_coords.add((_x + _x_step, _y - int(_y_step / 2)))
                            _unsure_coords.add((_x + _x_step, _y + int(_y_step / 2)))
                    else:
                        _grid[_x, _y] = _point_value

            while _unsure_coords:
                _current_coords = _unsure_coords.pop()
                if (_current_coords[0] < 0 or _current_coords[0] >= _target_points_per_axis or
                        _current_coords[1] < 0 or _current_coords[1] >= _target_points_per_axis):
                    continue
                if _current_coords in _confirmed_coords:
                    continue
                _x = _current_coords[0]
                _y = _current_coords[1]
                _local_constraints = [{'type': 'eq', 'fun': lambda x: x[_ids[0]] - _x_values[_x]},
                                      {'type': 'eq', 'fun': lambda x: x[_ids[1]] - _y_values[_y]}]
                _current_fun = self._calc_fun_with_constraints(_local_constraints)
                _grid_fun = _grid[_current_coords[0], _current_coords[1]]
                if ((_current_fun > _contour_fun and _grid_fun < _contour_fun) or
                        (_current_fun < _contour_fun and _grid_fun > _contour_fun)):
                    if _iterations % 2 == 0:
                        _unsure_coords.add((_x - _x_step, _y))
                        _unsure_coords.add((_x, _y - _y_step))
                        _unsure_coords.add((_x + _x_step, _y))
                        _unsure_coords.add((_x, _y + _y_step))
                    else:
                        _unsure_coords.add((_x - _x_step, _y - int(_y_step / 2)))
                        _unsure_coords.add((_x - _x_step, _y + int(_y_step / 2)))
                        _unsure_coords.add((_x + _x_step, _y - int(_y_step / 2)))
                        _unsure_coords.add((_x + _x_step, _y + int(_y_step / 2)))
                _grid[_current_coords[0], _current_coords[1]] = _current_fun
                _confirmed_coords.add(_current_coords)

            if _iterations % 2 == 0:
                _x_step = int(_x_step / 2)
            else:
                _y_step = int(_y_step / 2)
            _iterations += 1

        _left_cutoff = 0
        _right_cutoff = _target_points_per_axis - 1
        _bottom_cutoff = 0
        _top_cutoff = _target_points_per_axis - 1
        _padding = int(3 / area_scale_factor * max(1, 2 ** (iterations - 4)))

        while _right_cutoff > 0 and np.min(_grid[_right_cutoff]) > _contour_fun:
            _right_cutoff -= 1
        _right_cutoff += _padding
        _right_cutoff = min(_right_cutoff, _target_points_per_axis - 1)

        while _left_cutoff < _right_cutoff and np.min(_grid[_left_cutoff]) > _contour_fun:
            _left_cutoff += 1
        _left_cutoff -= _padding
        _left_cutoff = max(_left_cutoff, 0)

        _grid = _grid[_left_cutoff:_right_cutoff]
        _grid = _grid.T

        while _top_cutoff > 0 and np.min(_grid[_top_cutoff]) > _contour_fun:
            _top_cutoff -= 1
        _top_cutoff += _padding
        _top_cutoff = min(_top_cutoff, _target_points_per_axis - 1)

        while _bottom_cutoff < _top_cutoff and np.min(_grid[_bottom_cutoff]) > _contour_fun:
            _bottom_cutoff += 1
        _bottom_cutoff -= _padding
        _bottom_cutoff = max(_bottom_cutoff, 0)

        _grid = _grid[_bottom_cutoff:_top_cutoff]
        _grid = _grid.T

        _x_values = _x_values[_left_cutoff:_right_cutoff]
        _y_values = _y_values[_bottom_cutoff:_top_cutoff]

        _grid = np.sqrt(_grid - _min_fun)
        self._func_wrapper_unpack_args(self._par_val)
        return ContourFactory.create_grid_contour(_x_values, _y_values, _grid, sigma)

    @staticmethod
    def _heuristic_point_evaluation(contour_fun, grid, x, y, vector_1, vector_2):
        _adjacent_points = MinimizerScipyOptimize._get_adjacent_grid_points(grid, x, y, vector_1, vector_2)
        if np.max(_adjacent_points) < contour_fun:
            return np.mean(_adjacent_points)
        if np.min(_adjacent_points) > contour_fun:
            return np.mean(_adjacent_points)
        return -1

    @staticmethod
    def _get_adjacent_grid_points(grid, x_0, y_0, vector_1, vector_2):
        _x_size = np.ma.size(grid, 0)
        _y_size = np.ma.size(grid, 1)
        _grid_points = []
        for i in range(4):
            _x = x_0
            _y = y_0
            if i == 0:
                _x -= vector_1[0]
                _y -= vector_1[1]
            elif i == 1:
                _x -= vector_2[0]
                _y -= vector_2[1]
            elif i == 2:
                _x += vector_1[0]
                _y += vector_1[1]
            elif i == 3:
                _x += vector_2[0]
                _y += vector_2[1]
            if _x >= 0 and _x < _x_size and _y >= 0 and _y < _y_size:
                _grid_points.append(grid[_x][_y])
        return np.asarray(_grid_points)

    def _contour_beacon(self, parameter_name_1, parameter_name_2, sigma=1.0, beacon_size=0.02):

        _contour_fun = self.function_value + sigma ** 2
        _contour_fun_upper_tolerance = self.function_value + (1.2 * sigma) ** 2
        _contour_fun_lower_tolerance = self.function_value + (0.8 * sigma) ** 2
        _ids = (self._par_names.index(parameter_name_1), self._par_names.index(parameter_name_2))
        _minimum = np.asarray([self._par_val[_ids[0]], self._par_val[_ids[1]]])
        _err = np.asarray([self._par_err[_ids[0]], self._par_err[_ids[1]]])

        _angles = []

        CONTOUR_ELLIPSE_POINTS = 21
        CONTOUR_STRETCHING = 4.0
        _unstretched_angles = np.linspace(-np.pi / 2, np.pi / 2, CONTOUR_ELLIPSE_POINTS, endpoint=True)
        _contour_search_ellipse = np.empty((2, CONTOUR_ELLIPSE_POINTS))
        _contour_search_ellipse[0] = sigma * beacon_size * np.sin(_unstretched_angles)
        _contour_search_ellipse[1] = sigma * CONTOUR_STRETCHING * beacon_size * np.cos(_unstretched_angles)
        _stretched_absolute_angles = np.abs(np.arctan(np.tan(_unstretched_angles) / CONTOUR_STRETCHING))
        _curvature_adjustion_factors = 1 + 0.025 * (10 - _stretched_absolute_angles * 180 / np.pi)
        _curvature_adjustion_factors = np.where(_curvature_adjustion_factors >= 0.25, _curvature_adjustion_factors,
                                                0.25)

        _termination_distance = (sigma * CONTOUR_STRETCHING * beacon_size) ** 2

        _meta_cost_function = lambda z: (_contour_fun - self._calc_fun_with_constraints(
            [{'type': 'eq', 'fun': lambda x: x[_ids[0]] - (_minimum[0] + _err[0] * z)},
             {'type': 'eq', 'fun': lambda x: x[_ids[1]] - _minimum[1]}]))

        _start_x = opt.brentq(_meta_cost_function, 0, 2 * sigma, maxiter=1000)
        _start_point = np.asarray([_start_x, 0.0])

        _phi = self._calculate_tangential_angle(_start_point, _ids)
        _coords = _start_point
        _curvature_adjustion = 1.0
        _last_backtrack = 0

        _loops = 0

        _contour_coords = [_start_point]

        while (True):
            _transformed_search_ellipse = self._rotate_clockwise(_contour_search_ellipse * _curvature_adjustion, _phi)
            _transformed_search_ellipse[0] += _coords[0]
            _transformed_search_ellipse[1] += _coords[1]
            _transformed_search_ellipse = _transformed_search_ellipse.T
            _ellipse_fun_values = np.empty(CONTOUR_ELLIPSE_POINTS)
            for i in range(CONTOUR_ELLIPSE_POINTS):
                _ellipse_coords = _transformed_search_ellipse[i]
                _transformed_coords = self._transform_coordinates(_minimum, _ellipse_coords, _err)
                _point_constraints = [{"type": "eq", "fun": lambda x: x[_ids[0]] - _transformed_coords[0]},
                                      {"type": "eq", "fun": lambda x: x[_ids[1]] - _transformed_coords[1]}]
                _ellipse_fun_values[i] = self._calc_fun_with_constraints(_point_constraints)
            _min_index = np.argmin(np.abs(_ellipse_fun_values - _contour_fun))
            _new_coords = _transformed_search_ellipse[_min_index]

            _curvature_adjustion *= _curvature_adjustion_factors[_min_index]
            if _curvature_adjustion > 1.0:
                _curvature_adjustion = 1.0

            if _stretched_absolute_angles[_min_index] > 0.349111:
                print("BACKTRACK")
                _contour_coords = _contour_coords[0:-1]
            else:
                _contour_coords.append(_new_coords)

            # _delta = _contour_coords[-1] - _contour_coords[-2]
            # _phi = np.arctan2(_delta[0], _delta[1])
            _coords = _contour_coords[-1]
            _phi = self._calculate_tangential_angle(_coords, _ids)

            if np.sum((_coords - _start_point) ** 2) < _termination_distance and _loops > 10:
                break

            if _loops < 200:
                _loops += 1
            else:
                break
        self._func_wrapper_unpack_args(self._par_val)
        return ContourFactory.create_xy_contour(self._transform_contour(_minimum, _contour_coords, _err), sigma)

    @staticmethod
    def _transform_coordinates(minimum, sigma_coordinates, errors):
        return minimum + sigma_coordinates * errors

    @staticmethod
    def _transform_contour(minimum, sigma_contour, errors):
        _transformed_contour = []
        for _coords in sigma_contour:
            _transformed_contour.append(MinimizerScipyOptimize._transform_coordinates(minimum, _coords, errors))
        return _transformed_contour

    @staticmethod
    def _rotate_clockwise(xy_values, phi):
        _rotated_xy_values = np.empty(shape=xy_values.shape)
        _rotated_xy_values[0] = np.cos(phi) * xy_values[0] + np.sin(phi) * xy_values[1]
        _rotated_xy_values[1] = -np.sin(phi) * xy_values[0] + np.cos(phi) * xy_values[1]
        return _rotated_xy_values

    def _calculate_tangential_angle(self, coords, ids):
        _meta_cost_function_gradient = lambda pars: self._calc_fun_with_constraints(
            [{'type': 'eq', 'fun': lambda x: x[ids[0]] - (self._par_val[ids[0]] + self._par_err[ids[0]] * pars[0])},
             {'type': 'eq', 'fun': lambda x: x[ids[1]] - (self._par_val[ids[1]] + self._par_err[ids[1]] * pars[1])}])
        _grad = nd.Gradient(_meta_cost_function_gradient)(coords)
        return np.arctan2(_grad[0], _grad[1]) + np.pi / 2

    def _calc_fun_with_constraints(self, additional_constraints, continuous_x0=False):
        _local_constraints = self._par_constraints + additional_constraints
        if continuous_x0:
            _x0 = self._x0
        else:
            _x0 = self._par_val
        _result = opt.minimize(self._func_wrapper_unpack_args,
                               _x0,
                               args=(),
                               method="slsqp",
                               jac=None,
                               bounds=self._par_bounds,
                               constraints=_local_constraints,
                               tol=self.tolerance,
                               callback=None,
                               options=dict(maxiter=6000, disp=False))
        self._x0 = _result.x
        return _result.fun

    def profile(self, parameter_name, bins=21, bound=2, subtract_min=False):
        if not self.did_fit:
            raise RuntimeError("Need to perform a fit before calling profile()!")
        _par_id = self._par_names.index(parameter_name)
        _par_err = self.parameter_errors[_par_id]
        _par_min = self._par_val[_par_id]
        _par = np.linspace(
            start=_par_min - bound * _par_err,
            stop=_par_min + bound * _par_err,
            num=bins, endpoint=True
        )
        _y_offset = self.function_value if subtract_min else 0

        _y = np.empty(bins)
        self._x0 = self._par_val
        for i in range(bins):
            _y[i] = self._calc_fun_with_constraints(
                [{"type": "eq", "fun": lambda x: x[_par_id] - _par[i]}], continuous_x0=True
            )
        self._func_wrapper_unpack_args(self._par_val)
        return np.asarray([_par, _y - _y_offset])

    def _func_wrapper(self, *parameter_values):
        '''call FCN, but ensure fixed parameters are passed with their fixed value'''
        # Note: this is needed in order to ensure that derivatives of `_func_wrapper`
        #       take parameter fixing into account
        assert (len(self.parameter_values) == len(self._par_fixed) == len(parameter_values))
        # replace parameter values
        parameter_values = [
            _fixed_val if _is_fixed else _call_val
            for _call_val, _fixed_val, _is_fixed in zip(parameter_values, self.parameter_values, self._par_fixed)
        ]
        _res = MinimizerBase._func_wrapper(self, *parameter_values)
        # some scipy methods handle 'nan' incorrectly -> return MAX_FLOAT instead
        if np.isnan(_res):
            return np.finfo(float).max
        return _res



import unittest
from six import StringIO
import numpy as np

from kafe2.fit.io import IOStreamHandle
from kafe2.fit.representation import ConstraintYamlReader, ConstraintYamlWriter
from kafe2.core.constraint import GaussianSimpleParameterConstraint, \
    GaussianMatrixParameterConstraint

TEST_SIMPLE_GAUSSIAN_CONSTRAINT_ABS = """
type: simple
index: 3
value: 2.0
uncertainty: 1.2
"""

TEST_SIMPLE_GAUSSIAN_CONSTRAINT_REL = """
type: simple
index: 3
value: 2.0
uncertainty: 0.6
relative: True
"""

TEST_SIMPLE_GAUSSIAN_CONSTRAINT_MISSING_KEYWORD = """
type: simple
index: 3
uncertainty: 1.2
"""

TEST_SIMPLE_GAUSSIAN_CONSTRAINT_EXTRA_KEYWORD = TEST_SIMPLE_GAUSSIAN_CONSTRAINT_ABS + """
extra_keyword: 3.14
"""


class TestSimpleGaussianConstraintRepresenter(unittest.TestCase):

    def setUp(self):
        self._index = 3
        self._value = 2.0
        self._uncertainty = 1.2
        self._constraint = GaussianSimpleParameterConstraint(index=self._index, value=self._value,
                                                             uncertainty=self._uncertainty)

        self._roundtrip_stringstream = IOStreamHandle(StringIO())
        self._testfile_stringstream_abs = IOStreamHandle(StringIO(TEST_SIMPLE_GAUSSIAN_CONSTRAINT_ABS))
        self._testfile_stringstream_rel = IOStreamHandle(StringIO(TEST_SIMPLE_GAUSSIAN_CONSTRAINT_REL))

        self._roundtrip_streamreader = ConstraintYamlReader(self._roundtrip_stringstream)
        self._roundtrip_streamwriter = ConstraintYamlWriter(self._constraint, self._roundtrip_stringstream)
        self._testfile_streamreader_abs = ConstraintYamlReader(self._testfile_stringstream_abs)
        self._testfile_streamreader_rel = ConstraintYamlReader(self._testfile_stringstream_rel)

        self._testfile_stringstream_missing_keyword = IOStreamHandle(
            StringIO(TEST_SIMPLE_GAUSSIAN_CONSTRAINT_MISSING_KEYWORD))
        self._testfile_stringstream_extra_keyword = IOStreamHandle(
            StringIO(TEST_SIMPLE_GAUSSIAN_CONSTRAINT_EXTRA_KEYWORD))
        self._testfile_streamreader_missing_keyword = ConstraintYamlReader(
            self._testfile_stringstream_missing_keyword)
        self._testfile_streamreader_extra_keyword = ConstraintYamlReader(
            self._testfile_stringstream_extra_keyword)

    def _assert_constraints_equal(self, constraint_1, constraint_2):
        self.assertTrue(constraint_1.index == constraint_2.index)
        self.assertTrue(constraint_1.value == constraint_2.value)
        self.assertTrue(constraint_1.uncertainty == constraint_2.uncertainty)
        self.assertTrue(constraint_1.uncertainty_rel == constraint_2.uncertainty_rel)

    def test_write_to_roundtrip_stringstream(self):
        self._roundtrip_streamwriter.write()

    def test_read_from_testfile_stream(self):
        _read_constraint = self._testfile_streamreader_abs.read()
        self.assertTrue(isinstance(_read_constraint, GaussianSimpleParameterConstraint))
        self._assert_constraints_equal(_read_constraint, self._constraint)
        _read_constraint = self._testfile_streamreader_rel.read()
        self.assertTrue(isinstance(_read_constraint, GaussianSimpleParameterConstraint))
        self._assert_constraints_equal(_read_constraint, self._constraint)

    def test_read_from_testfile_stream_missing_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_missing_keyword.read()

    def test_read_from_testfile_stream_extra_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_extra_keyword.read()

    def test_round_trip_with_stringstream(self):
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_constraint = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_constraint, GaussianSimpleParameterConstraint))
        self._assert_constraints_equal(_read_constraint, self._constraint)


TEST_MATRIX_GAUSSIAN_CONSTRAINT_COV_ABS = """
type: matrix
indices: [3, 0, 1]
values: [1.0, 10.0, 100.0]
matrix: [[0.01, 0.01, 0.2], [0.01, 1.0, 3.0], [0.2, 3.0, 100.0]]
"""

TEST_MATRIX_GAUSSIAN_CONSTRAINT_COV_REL = """
type: matrix
indices: [3, 0, 1]
values: [1.0, 10.0, 100.0]
matrix: [[0.01, 0.001, 0.002], [0.001, 0.01, 0.003], [0.002, 0.003, 0.01]]
relative: True
"""

TEST_MATRIX_GAUSSIAN_CONSTRAINT_COR_ABS = """
type: matrix
indices: [3, 0, 1]
values: [1.0, 10.0, 100.0]
matrix: [[1.0, 0.1, 0.2], [0.1, 1.0, 0.3], [0.2, 0.3, 1.0]]
matrix_type: cor
uncertainties: [0.1, 1.0, 10.0]
"""

TEST_MATRIX_GAUSSIAN_CONSTRAINT_COR_REL = """
type: matrix
indices: [3, 0, 1]
values: [1.0, 10.0, 100.0]
matrix: [[1.0, 0.1, 0.2], [0.1, 1.0, 0.3], [0.2, 0.3, 1.0]]
matrix_type: cor
uncertainties: [0.1, 0.1, 0.1]
relative: True
"""

TEST_MATRIX_GAUSSIAN_CONSTRAINT_MISSING_KEYWORD = """
type: matrix
values: [1.0, 10.0, 100.0]
matrix: [[0.1, 0.1, 2.0], [0.1, 10.0, 30.0], [2.0, 30.0, 1000.0]]
"""

TEST_MATRIX_GAUSSIAN_CONSTRAINT_EXTRA_KEYWORD = TEST_MATRIX_GAUSSIAN_CONSTRAINT_COV_ABS + """
extra_keyword: 3.14
"""


class TestMatrixGaussianConstraintRepresenter(unittest.TestCase):

    def setUp(self):
        self._indices = [3, 0, 1]
        self._values = [1.0, 10.0, 100.0]
        self._cov_mat = [[0.01, 0.01, 0.2], [0.01, 1.0, 3.0], [0.2, 3.0, 100.0]]
        self._constraint = GaussianMatrixParameterConstraint(indices=self._indices, values=self._values,
                                                             matrix=self._cov_mat)

        self._roundtrip_stringstream = IOStreamHandle(StringIO())
        self._testfile_stringstream_cov_abs = IOStreamHandle(StringIO(TEST_MATRIX_GAUSSIAN_CONSTRAINT_COV_ABS))
        self._testfile_stringstream_cov_rel = IOStreamHandle(StringIO(TEST_MATRIX_GAUSSIAN_CONSTRAINT_COV_REL))
        self._testfile_stringstream_cor_abs = IOStreamHandle(StringIO(TEST_MATRIX_GAUSSIAN_CONSTRAINT_COR_ABS))
        self._testfile_stringstream_cor_rel = IOStreamHandle(StringIO(TEST_MATRIX_GAUSSIAN_CONSTRAINT_COR_REL))

        self._roundtrip_streamreader = ConstraintYamlReader(self._roundtrip_stringstream)
        self._roundtrip_streamwriter = ConstraintYamlWriter(self._constraint, self._roundtrip_stringstream)
        self._testfile_streamreader_cov_abs = ConstraintYamlReader(self._testfile_stringstream_cov_abs)
        self._testfile_streamreader_cov_rel = ConstraintYamlReader(self._testfile_stringstream_cov_rel)
        self._testfile_streamreader_cor_abs = ConstraintYamlReader(self._testfile_stringstream_cor_abs)
        self._testfile_streamreader_cor_rel = ConstraintYamlReader(self._testfile_stringstream_cor_rel)

        self._testfile_stringstream_missing_keyword = IOStreamHandle(
            StringIO(TEST_MATRIX_GAUSSIAN_CONSTRAINT_MISSING_KEYWORD))
        self._testfile_stringstream_extra_keyword = IOStreamHandle(
            StringIO(TEST_MATRIX_GAUSSIAN_CONSTRAINT_EXTRA_KEYWORD))
        self._testfile_streamreader_missing_keyword = ConstraintYamlReader(
            self._testfile_stringstream_missing_keyword)
        self._testfile_streamreader_extra_keyword = ConstraintYamlReader(
            self._testfile_stringstream_extra_keyword)

    def _assert_constraints_equal(self, constraint_1, constraint_2):
        self.assertTrue(np.all(constraint_1.indices == constraint_2.indices))
        self.assertTrue(np.all(constraint_1.values == constraint_2.values))
        self.assertTrue(np.allclose(constraint_1.cov_mat, constraint_2.cov_mat))
        self.assertTrue(np.allclose(constraint_1.cov_mat_rel, constraint_2.cov_mat_rel))
        self.assertTrue(np.allclose(constraint_1.cor_mat, constraint_2.cor_mat))
        self.assertTrue(np.allclose(constraint_1.uncertainties, constraint_2.uncertainties))
        self.assertTrue(np.allclose(constraint_1.uncertainties_rel, constraint_2.uncertainties_rel))

    def test_write_to_roundtrip_stringstream(self):
        self._roundtrip_streamwriter.write()

    def test_read_from_testfile_stream(self):
        _read_constraint = self._testfile_streamreader_cov_abs.read()
        self.assertTrue(isinstance(_read_constraint, GaussianMatrixParameterConstraint))
        self._assert_constraints_equal(_read_constraint, self._constraint)
        _read_constraint = self._testfile_streamreader_cov_rel.read()
        self.assertTrue(isinstance(_read_constraint, GaussianMatrixParameterConstraint))
        self._assert_constraints_equal(_read_constraint, self._constraint)
        _read_constraint = self._testfile_streamreader_cor_abs.read()
        self.assertTrue(isinstance(_read_constraint, GaussianMatrixParameterConstraint))
        self._assert_constraints_equal(_read_constraint, self._constraint)
        _read_constraint = self._testfile_streamreader_cor_rel.read()
        self.assertTrue(isinstance(_read_constraint, GaussianMatrixParameterConstraint))
        self._assert_constraints_equal(_read_constraint, self._constraint)

    def test_read_from_testfile_stream_missing_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_missing_keyword.read()

    def test_read_from_testfile_stream_extra_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_extra_keyword.read()

    def test_round_trip_with_stringstream(self):
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_constraint = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_constraint, GaussianMatrixParameterConstraint))
        self._assert_constraints_equal(_read_constraint, self._constraint)



import numpy as np
import six
import warnings

from scipy.stats import poisson, norm, chi2
from scipy.linalg import solve_triangular
from ..io.file import FileIOMixin
from .format import ParameterFormatter, CostFunctionFormatter
from ..util import cholesky_decomposition, log_determinant

if six.PY2:
    from funcsigs import signature
else:
    from inspect import signature


__all__ = ["CostFunction", "CostFunction_Chi2", "CostFunction_NegLogLikelihood",
           "CostFunction_GaussApproximation"]


class CostFunction(FileIOMixin, object):
    """
    Base class for cost functions. Built from a Python function with some extra functionality used
    by Fit objects.

    Any Python function returning a ``float`` can be used as a cost function,
    although a number of common cost functions are provided as built-ins for
    all fit types.

    In order to be used as a model function, a native Python function must be wrapped
    by an object whose class derives from this base class.
    There is a dedicated :py:class:`CostFunction` specialization for each type of
    fit.

    This class provides the basic functionality used by all :py:class:`CostFunction` objects.
    These use introspection (:py:mod:`inspect`) for determining the parameter structure of the
    cost function and to ensure the function can be used as a cost function (validation).
    """

    _DATA_NAME = "data"
    _MODEL_NAME = "model"
    _COV_MAT_CHOLESKY_NAME = "total_cov_mat_cholesky"
    _ERROR_NAME = "total_error"

    def __init__(
            self, cost_function, arg_names=None, add_constraint_cost=True,
            add_determinant_cost=False):
        """
        Construct :py:class:`CostFunction` object (a wrapper for a native Python function):

        :param typing.Callable cost_function: function handle
        :param typing.Iterable[str] arg_names: the names to use for the cost function arguments.
            If None, detect from function signature.
        :param bool add_constraint_cost: If :py:obj:`True`, automatically add the cost for kafe2
            constraints.
        :param bool add_determinant_cost: If :py:obj:`True`, automatically increase the cost
            function value by the logarithm of the determinant of the covariance matrix to reduce
            bias.
        """
        self._cost_function_handle = cost_function
        _signature = signature(self._cost_function_handle)
        if arg_names is None:
            self._arg_names = list(_signature.parameters.keys())
            for _par in _signature.parameters.values():
                if _par.kind == _par.VAR_POSITIONAL:
                    raise ValueError(
                        "Cost function '{}' with variable number of positional arguments "
                        "(*{}) needs explicit argument names.".format(
                            self._cost_function_handle.__name__, _par.name))
        else:
            self._arg_names = list(arg_names)
        if "cost" in self._arg_names:
            raise ValueError(
                "The alias 'cost' for the cost function value cannot be used as a name for one of"
                "the cost function arguments!")
        for _par in _signature.parameters.values():
            if _par.kind == _par.VAR_KEYWORD:
                raise ValueError("Cost function '{}' with variable number of keyword arguments "
                                 "(**{}) is not supported.".format(
                                     self._cost_function_handle.__name__, _par.name, ))
        self._arg_count = len(self._arg_names)
        self._formatter = CostFunctionFormatter(
            name=self.name,
            arg_formatters=[ParameterFormatter(_arg_name, value=0, error=None)
                            for _arg_name in self._arg_names])

        self._add_constraint_cost = add_constraint_cost
        if self._add_constraint_cost:
            self._arg_names += ["parameter_values", "parameter_constraints"]
            self._arg_count += 2
        self._add_determinant_cost = add_determinant_cost
        if self._add_determinant_cost:
            if self.pointwise:
                self._arg_names += ["total_error_squared_log_sum"]
            else:
                self._arg_names += ["total_cov_mat_log_determinant"]
            self._arg_count += 1

        self._errors_valid = True
        self._needs_errors = True
        self._is_chi2 = False
        self._saturated = False
        self._kafe2go_identifier = None
        super(CostFunction, self).__init__()

    @classmethod
    def _get_base_class(cls):
        return CostFunction

    @classmethod
    def _get_object_type_name(cls):
        return 'cost_function'

    def __call__(self, *args):
        additional_cost = 0.0
        if self._add_determinant_cost:
            _log_determinant = args[-1]
            args = args[:-1]
            if _log_determinant is not None:
                additional_cost += _log_determinant
        if self._add_constraint_cost:
            _par_constraints = args[-1]
            _par_vals = args[-2]
            args = args[:-2]
            if _par_constraints is not None:
                for _par_constraint in _par_constraints:
                    additional_cost += _par_constraint.cost(_par_vals)
        return self._cost_function_handle(*args) + additional_cost

    @property
    def name(self):
        """The cost function name (a valid Python identifier)"""
        return self._cost_function_handle.__name__

    @property
    def func(self):
        """The cost function handle"""
        return self._cost_function_handle

    @property
    def arg_names(self):
        """The names of the cost function arguments."""
        return self._arg_names

    @property
    def formatter(self):
        """The :py:obj:`Formatter` object for this function"""
        return self._formatter

    @property
    def argument_formatters(self):
        """The :py:obj:`Formatter` objects for the function arguments"""
        return self._formatter.arg_formatters

    @property
    def needs_errors(self):
        """Whether the cost function needs errors for a meaningful result"""
        return self._needs_errors

    @property
    def is_chi2(self):
        """Whether the cost function is a chi2 cost function."""
        return self._is_chi2

    @property
    def saturated(self):
        """Whether the cost function value is calculated from a saturated likelihood."""
        return self._saturated

    @property
    def add_determinant_cost(self):
        """Whether the determinant cost is being added automatically to the cost function value."""
        return self._add_determinant_cost

    @property
    def kafe2go_identifier(self):
        """Short string representation (if any) of this cost function when dumping to file."""
        return self._kafe2go_identifier

    @property
    def pointwise(self):
        """True if cost function result does not depend on covariances."""
        return False

    @property
    def pointwise_version(self):
        """Optimized version of cost function that uses pointwise errors, can be None."""
        return None

    @property
    def errors_valid(self):
        return self._errors_valid

    def goodness_of_fit(self, *args):
        """How well the model agrees with the data."""
        try:
            _index_data = self._arg_names.index(self._DATA_NAME)
            _index_model = self._arg_names.index(self._MODEL_NAME)
        except ValueError:
            return None
        if self._add_determinant_cost:
            args = args[:-1] + (0.0,)
        _cost = self(*args)
        args = list(args)
        if self._add_determinant_cost:
            args = args[:-1]
        if self._add_constraint_cost:
            args = args[:-2]
        args[_index_model] = args[_index_data]
        _saturated_cost = self._cost_function_handle(*args)
        return _cost - _saturated_cost

    def chi2_probability(self, cost_function_value, ndf):
        """The chi2 probability associated with this cost function, None for non-chi2 cost
        functions.

        :param float cost_function_value: the associated cost function value.
        :param int ndf: the associated number of degrees of freedom.
        :returns: the associated chi2 probability.
        :rtype: float or None
        """
        return 1.0 - chi2.cdf(cost_function_value, ndf) if self.is_chi2 else None

    def get_uncertainty_gaussian_approximation(self, data):
        """Get the gaussian approximation of the uncertainty inherent to the cost function, returns
        0 by default.

        :param data: the fit data
        :return: the approximated gaussian uncertainty given the fit data
        """
        return 0

    def is_data_compatible(self, data):
        """Tests if model data is compatible with cost function

        :param data: the fit data
        :type data: numpy.ndarray
        :return: if the data is compatible, and if not a reason for the incompatibility
        :rtype: (boo, str)
        """
        return True, None


class CostFunction_Chi2(CostFunction):
    def __init__(
            self, errors_to_use='covariance', fallback_on_singular=True, add_constraint_cost=True,
            add_determinant_cost=True):
        """Base class for built-in least-squares cost function.

        :param errors_to_use: Which errors to use when calculating :math:`\\chi^2`.
            Either ``'covariance'``, ``'pointwise'`` or ``None``.
        :type errors_to_use: str or None
        :param bool fallback_on_singular: If :py:obj:`True` and the covariance matrix is singular
            (or the errors are zero), calculate :math:`\\chi^2` as with ``errors_to_use=None``
        :param bool add_constraint_cost: If :py:obj:`True`, automatically add the cost for kafe2
            constraints.
        :param bool add_determinant_cost: If :py:obj:`True`, automatically increase the cost
            function value by the logarithm of the determinant of the covariance matrix to reduce
            bias.
        """

        _cost_function_description = "chi-square"
        if errors_to_use is None:
            _chi2_func = self.chi2_no_errors
            _arg_names = [self._DATA_NAME, self._MODEL_NAME]
            self._fail_on_no_matrix = False
            self._fail_on_no_errors = False
            _cost_function_description += " (no uncertainties)"
        elif errors_to_use.lower() == 'covariance':
            _chi2_func = self.chi2_covariance
            _arg_names = [self._DATA_NAME, self._MODEL_NAME, self._COV_MAT_CHOLESKY_NAME]
            self._fail_on_no_matrix = not fallback_on_singular
            self._fail_on_no_errors = True
            _cost_function_description += " (with covariance matrix)"
        elif errors_to_use.lower() == 'pointwise':
            _chi2_func = self.chi2_pointwise_errors
            _arg_names = [self._DATA_NAME, self._MODEL_NAME, self._ERROR_NAME]
            self._fail_on_no_matrix = False
            self._fail_on_no_errors = not fallback_on_singular
            _cost_function_description += " (with pointwise errors)"
        else:
            raise ValueError(
                "Unknown value '%s' for 'errors_to_use': must be one of "
                "('covariance', 'pointwise', None)")

        super(CostFunction_Chi2, self).__init__(
            cost_function=_chi2_func,
            arg_names=_arg_names,
            add_constraint_cost=add_constraint_cost,
            add_determinant_cost=add_determinant_cost
        )

        self._formatter.latex_name = "\\chi^2"
        self._formatter.name = "chi2"
        self._formatter.description = _cost_function_description
        self._errors_valid = self._needs_errors = errors_to_use is not None
        self._is_chi2 = True
        self._saturated = True
        self._kafe2go_identifier = self.name

    def _chi2(self, data, model, cov_mat_cholesky=None, err=None):
        data = np.asarray(data)
        model = np.asarray(model)

        if model.shape != data.shape:
            raise ValueError(
                "'data' and 'model' must have the same shape! Got %r and %r..."
                % (data.shape, model.shape))

        _res = (data - model)

        # if a covariance matrix inverse is given, use it
        if cov_mat_cholesky is not None:
            try:
                _x = solve_triangular(cov_mat_cholesky, _res, lower=True)
            except ValueError:
                return np.inf
            return np.inner(_x, _x)

        if self._fail_on_no_matrix:
            raise ValueError("Covariance matrix is singular!")

        # otherwise, if an array of point-wise errors is given, use that
        if err is not None:
            if np.any(err == 0.0):
                if self._fail_on_no_errors:
                    raise ValueError("'err' must not contain any zero values!")
                if self.needs_errors:
                    # There are other warnings that notify the user about singular cov mat, etc.
                    warnings.warn("Setting all data errors to 1 as a fallback.")
            else:
                _res = _res / err

        # return sum of squared residuals
        _cost = np.inner(_res, _res)
        if np.isnan(_cost):
            _cost = np.inf
        return _cost

    def chi2_no_errors(self, data, model):
        r"""A least-squares cost function calculated from (`y`) data and model values,
        without considering uncertainties:

        .. math::
            C = \chi^2({\bf d}, {\bf m}) = \sum_k (d_k - m_k)^2
                +
                C_{\rm con}({\bf p}).

        In the above, :math:`{\bf d}` are the measurements,
        :math:`{\bf m}` are the model predictions,
        and :math:`C_{\rm con}({\bf p})` is the additional cost resulting from any constrained
        parameters.

        :param data: measurement data :math:`{\bf d}`
        :param model: model predictions :math:`{\bf m}`

        :return: cost function value
        """
        return self._chi2(data=data, model=model)

    def chi2_covariance(self, data, model, total_cov_mat_cholesky):
        r"""A least-squares cost function calculated from (`y`) data and model values,
        considering the covariance matrix of the (`y`) measurements.
        The cost function value can be calculated as follows:

        .. math::
            C = \chi^2({\bf d}, {\bf m})
            = ({\bf d} - {\bf m})^{\top}\,{{\bf V}^{-1}}\,({\bf d} - {\bf m})
                +
                C_{\rm con}({\bf p})
                +
                C_{\rm det}({\bf V}).

        In the above, :math:`{\bf d}` are the measurements,
        :math:`{\bf m}` are the model predictions,
        :math:`{{\bf V}^{-1}}` is the inverse of the total covariance matrix,
        :math:`C_{\rm con}({\bf p})` is the additional cost resulting from any constrained
        parameters,
        and :math:`C_{\rm det}({\bf V}) = \ln \det({\bf V})` is the additional cost to compensate
        for a non-constant covariance matrix.

        :param data: measurement data :math:`{\bf d}`
        :param model: model predictions :math:`{\bf m}`
        :param total_cov_mat_cholesky: Cholesky decomposition  of the total covariance matrix
            :math:`{\bf L}` with :math:`{\bf L}^\top {\bf L} = {\bf V}`

        :return: cost function value
        """
        return self._chi2(data=data, model=model, cov_mat_cholesky=total_cov_mat_cholesky)

    def chi2_pointwise_errors(self, data, model, total_error):
        r"""A least-squares cost function calculated from (`y`) data and model values,
        considering pointwise (uncorrelated) uncertainties for each data point:

        .. math::
            C = \chi^2({\bf d}, {\bf m}, {\bf \sigma}) = \sum_k \frac{d_k - m_k}{\sigma_k}
                +
                C_{\rm con}({\bf p})
                +
                C_{\rm det}({\bf \sigma}).

        In the above, :math:`{\bf d}` are the measurements,
        :math:`{\bf m}` are the model predictions,
        :math:`{\bf \sigma}` are the pointwise total uncertainties,
        :math:`C_{\rm con}({\bf p})` is the additional cost resulting from any constrained
        parameters,
        and :math:`C_{\rm det}({\bf \sigma}) = \ln \prod_k \sigma_k^2` is the additional cost
        to compensate for non-constant errors.

        :param data: measurement data :math:`{\bf d}`
        :param model: model predictions :math:`{\bf m}`
        :param total_error: total error vector :math:`{\bf \sigma}`

        :return: cost function value
        """
        return self._chi2(data=data, model=model, err=total_error)

    @property
    def pointwise(self):
        return self._cost_function_handle != self.chi2_covariance

    @property
    def pointwise_version(self):
        if self._cost_function_handle == self.chi2_covariance:
            return type(self)(
                errors_to_use="pointwise",
                fallback_on_singular=not self._fail_on_no_matrix,
                add_constraint_cost=self._add_constraint_cost,
                add_determinant_cost=self._add_determinant_cost
            )
        else:
            return None


class CostFunction_NegLogLikelihood(CostFunction):
    def __init__(self, data_point_distribution='poisson', ratio=False):
        r"""
        Base class for built-in negative log-likelihood cost function.

        In addition to the measurement data and model predictions, likelihood-fits require a
        probability distribution describing how the measurements are distributed around the model
        predictions.
        This built-in cost function supports two such distributions: the *Poisson* and *Gaussian*
        (normal) distributions.

        In general, a negative log-likelihood cost function is defined as the double negative
        logarithm of the product of the individual likelihoods of the data points.

        The likelihood ratio is defined as ratio of the likelihood function for each individual
        observation, divided by the so-called *marginal likelihood*.

        :param data_point_distribution: Which type of statistics to use for modelling the
            distribution of individual data points. Either ``'poisson'`` or ``'gaussian'``.
        :type data_point_distribution: str
        :param ratio: If :py:obj:`True`, divide the likelihood by the marginal likelihood.
        :type ratio: bool
        """

        _cost_function_description = "negative log-likelihood"
        if data_point_distribution.lower() == "gaussian":
            if ratio:
                _nll_func = self.nllr_gaussian
                _cost_function_description += " ratio"
            else:
                _nll_func = self.nll_gaussian
            _cost_function_description += " (Gaussian uncertainties)"
            _arg_names = [self._DATA_NAME, self._MODEL_NAME, self._ERROR_NAME]
        elif data_point_distribution.lower() == "poisson":
            if ratio:
                _nll_func = self.nllr_poisson
                _cost_function_description += " ratio"
            else:
                _nll_func = self.nll_poisson
            _cost_function_description += " (Poisson uncertainties)"
            _arg_names = [self._DATA_NAME, self._MODEL_NAME]
        else:
            raise ValueError("Unknown value '%s' for 'data_point_distribution': "
                             "must be one of ('gaussian', 'poisson')!")

        super(CostFunction_NegLogLikelihood, self).__init__(
            cost_function=_nll_func,
            arg_names=_arg_names,
            add_determinant_cost=False
        )

        if ratio:
            self._formatter.latex_name = r"-2\ln\mathcal{L}_{\rm R}"
            self._formatter.name = "nllr"
        else:
            self._formatter.latex_name = r"-2\ln\mathcal{L}"
            self._formatter.latex_name_saturated = r"-2\ln\mathcal{L}_{\rm R}"
            self._formatter.name = "nll"
            self._formatter.name_saturated = "nllr"
        self._formatter.description = _cost_function_description
        self._needs_errors = _nll_func in [self.nll_gaussian, self.nllr_gaussian]
        self._saturated = ratio
        self._kafe2go_identifier = self.name

    @staticmethod
    def nll_gaussian(data, model, total_error):
        r"""A negative log-likelihood function assuming Gaussian statistics for each measurement.

        The cost function is given by:

        .. math::
            C = -2 \ln \mathcal{L}({\bf d}, {\bf m}, {\bf \sigma})
              = -2 \ln \prod_j \mathcal{L}_{\rm Gaussian} (x=d_j, \mu=m_j, \sigma=\sigma_j)
                + C_{\rm con}({\bf p}).

        .. math::
            \rightarrow C = -2 \ln \prod_j \frac{1}{\sqrt{2{\sigma_j}^2\pi}}
                            \exp{\left(-\frac{ (d_j-m_j)^2 }{ {\sigma_j}^2}\right)}
                            +
                            C_{\rm con}({\bf p}).

        In the above, :math:`{\bf d}` are the measurements,
        :math:`{\bf m}` are the model predictions,
        :math:`{\bf \sigma}` are the pointwise total uncertainties,
        and :math:`C_{\rm con}({\bf p})` is the additional cost resulting from any constrained
        parameters.

        :param data: measurement data :math:`{\bf d}`
        :param model: model predictions :math:`{\bf m}`
        :param total_error: total error vector :math:`{\bf \sigma}`

        :return: cost function value
        """

        _total_log_likelihood = np.sum(norm.logpdf(data, loc=model, scale=total_error))
        # guard against returning NaN
        if np.isnan(_total_log_likelihood):
            return np.inf
        return -2.0 * _total_log_likelihood

    @staticmethod
    def nll_poisson(data, model):
        r"""A negative log-likelihood function assuming Poisson statistics for each measurement.

        The cost function is given by:

        .. math::
            C = -2 \ln \mathcal{L}({\bf d}, {\bf m})
              = -2 \ln \prod_j \mathcal{L}_{\rm Poisson} (k=d_j, \lambda=m_j)
                +
                C_{\rm con}({\bf p}).

        .. math::
            \rightarrow C = -2 \ln \prod_j \frac{{m_j}^{d_j} \exp(-m_j)}{d_j!}
                            +
                            C_{\rm con}({\bf p}).

        In the above, :math:`{\bf d}` are the measurements,
        :math:`{\bf m}` are the model predictions,
        and :math:`C_{\rm con}({\bf p})` is the additional cost resulting from any constrained
        parameters.

        :param data: measurement data :math:`{\bf d}`
        :param model: model predictions :math:`{\bf m}`

        :return: cost function value
        """

        _total_log_likelihood = np.sum(poisson.logpmf(data, mu=model, loc=0.0))
        # guard against returning NaN
        if np.isnan(_total_log_likelihood):
            return np.inf
        return -2.0 * _total_log_likelihood

    @staticmethod
    def nllr_gaussian(data, model, total_error):
        _total_log_likelihood = np.sum(norm.logpdf(data, loc=model, scale=total_error))
        _saturated_log_likelihood = np.sum(norm.logpdf(x=data, loc=data, scale=total_error))
        _log_likelihood_ratio = _total_log_likelihood - _saturated_log_likelihood
        # guard against returning NaN
        if np.isnan(_log_likelihood_ratio):
            return np.inf
        return -2.0 * _log_likelihood_ratio

    @staticmethod
    def nllr_poisson(data, model):
        _total_log_likelihood = np.sum(poisson.logpmf(k=data, mu=model, loc=0.0))
        _saturated_log_likelihood = np.sum(poisson.logpmf(k=data, mu=data, loc=0.0))
        _log_likelihood_ratio = _total_log_likelihood - _saturated_log_likelihood
        # guard against returning NaN
        if np.isnan(_log_likelihood_ratio):
            return np.inf
        return -2.0 * _log_likelihood_ratio

    def is_data_compatible(self, data):
        if self._cost_function_handle in [self.nll_poisson, self.nllr_poisson] \
                and (np.count_nonzero(data % 1) > 0 or np.any(data < 0)):
            return False, "poisson distribution can only have non-negative integers as y data."
        return True, None

    def get_uncertainty_gaussian_approximation(self, data):
        if self._cost_function_handle in [self.nll_poisson, self.nllr_poisson]:
            return np.sqrt(data)
        return 0


class CostFunction_GaussApproximation(CostFunction):
    def __init__(
            self, errors_to_use='covariance', add_constraint_cost=True, add_determinant_cost=True):
        """
        Base class for built-in Gaussian approximation of the Poisson negative log-likelihood cost
        function.

        :param errors_to_use: Which errors to use when calculating :math:`\\chi^2`.
            Either ``'covariance'``, ``'pointwise'``.
        :type errors_to_use: str
        :param bool add_constraint_cost: If :py:obj:`True`, automatically add the cost for kafe2
            constraints.
        :param bool add_determinant_cost: If :py:obj:`True`, automatically increase the cost
            function value by the logarithm of the determinant of the covariance matrix to reduce
            bias.
        """

        _cost_function_description = "Gaussian approximation of Poisson NLL"
        if errors_to_use.lower() == 'covariance':
            _cost_function = self.gaussian_approximation_covariance
            _arg_names = [self._DATA_NAME, self._MODEL_NAME, self._COV_MAT_CHOLESKY_NAME]
            _cost_function_description += " (with covariance matrix)"
        elif errors_to_use.lower() == 'pointwise':
            _cost_function = self.gaussian_approximation_pointwise_errors
            _arg_names = [self._DATA_NAME, self._MODEL_NAME, self._ERROR_NAME]
            _cost_function_description += " (with pointwise errors)"
        else:
            raise ValueError(
                "Unknown value '%s' for 'errors_to_use': must be one of "
                "('covariance', 'pointwise')")

        super().__init__(
            cost_function=_cost_function,
            arg_names=_arg_names,
            add_constraint_cost=add_constraint_cost,
            add_determinant_cost=False
        )
        self._add_determinant_cost_ga = add_determinant_cost

        self._formatter.latex_name = r"-2\ln\tilde{\mathcal{L}}"
        self._formatter.latex_name_saturated = r"-2\ln\tilde{\mathcal{L}}_{\rm R}"
        self._formatter.description = _cost_function_description
        self._needs_errors = False
        self._is_chi2 = False
        self._saturated = True
        self._kafe2go_identifier = self.name

    def gaussian_approximation_covariance(self, data, model, total_cov_mat):
        r"""A least-squares cost function calculated from (`y`) data and model values,
        considering the covariance matrix of the (`y`) measurements.
        The cost function value can be calculated as follows:

        .. math::
            C = \chi^2({\bf d}, {\bf m}, {\bf V})
            = ({\bf d} - {\bf m})^{\top}\,{\tilde{\bf V}^{-1}}\,({\bf d} - {\bf m})
                +
                C_{\rm con}({\bf p})
                +
                C_{\rm det}(\tilde{\bf V}); \quad
            \tilde{{\bf V}}_{ij} = {\bf V}_{ij} + \delta_{ij} {\bf m}_i.

        In the above, :math:`{\bf d}` are the measurements,
        :math:`{\bf m}` are the model predictions,
        :math:`{{\bf V}^{-1}}` is the inverse of the total covariance matrix,
        :math:`C_{\rm con}({\bf p})` is the additional cost resulting from any constrained
        parameters,
        and :math:`C_{\rm det}(\tilde{\bf V}) = \ln \det(\tilde{\bf V})` is the additional cost to
        compensate for a non-constant covariance matrix.

        :param data: measurement data :math:`{\bf d}`
        :param model: model predictions :math:`{\bf m}`
        :param total_cov_mat: The total covariance matrix :math:`{\bf V}`

        :return: cost function value
        """
        _cholesky = cholesky_decomposition(total_cov_mat + np.diag(model))
        _residuals = model - data
        try:
            _x = solve_triangular(_cholesky, _residuals, lower=True)
        except ValueError:
            return np.inf
        _cost = np.sum(np.square(_x))
        if self._add_determinant_cost_ga:
            _cost += log_determinant(_cholesky)
        return _cost

    def gaussian_approximation_pointwise_errors(self, data, model, total_error):
        r"""A least-squares cost function calculated from data and model values,
        considering pointwise (uncorrelated) uncertainties for each data point:

        .. math::
            C = \chi^2({\bf d}, {\bf m}, {\bf \sigma})
                = \sum_k \left( \frac{d_k - m_k}{\sigma_k + \sqrt{m_k}} \right)^2
                +
                C_{\rm con}({\bf p})
                +
                C_{\rm det}({\bf \sigma}).

        In the above, :math:`{\bf d}` are the measurements,
        :math:`{\bf m}` are the model predictions,
        :math:`{\bf \sigma}` are the pointwise total uncertainties,
        :math:`C({\bf p})` is the additional cost resulting from any constrained parameters,
        and :math:`C_{\rm det}({\bf \sigma}) = \ln \prod_k m_k + \sigma_k^2` is the additional cost
        to compensate for non-constant errors.

        :param data: measurement data :math:`{\bf d}`
        :param model: model predictions :math:`{\bf m}`
        :param total_error: total error vector :math:`{\bf \sigma}`

        :return: cost function value
        """
        _residuals = model - data
        if np.all(_residuals == 0):
            return 0
        _variances = model + total_error ** 2
        _cost = np.sum(np.square(_residuals) / _variances)
        if self._add_determinant_cost_ga:
            _cost += np.sum(np.log(_variances))
        if np.isnan(_cost):
            _cost = np.inf
        return _cost

    @property
    def pointwise(self):
        return self._cost_function_handle == self.gaussian_approximation_pointwise_errors

    @property
    def pointwise_version(self):
        if self._cost_function_handle == self.gaussian_approximation_covariance:
            return type(self)(
                errors_to_use="pointwise",
                add_constraint_cost=self._add_constraint_cost,
                add_determinant_cost=self._add_determinant_cost
            )
        else:
            return None

    def goodness_of_fit(self, *args):
        _original_add_determinant_cost_pp = self._add_determinant_cost_ga
        self._add_determinant_cost_ga = False
        _gof = super().goodness_of_fit(*args)
        self._add_determinant_cost_ga = _original_add_determinant_cost_pp
        return _gof

    def get_uncertainty_gaussian_approximation(self, data):
        return np.sqrt(data)


STRING_TO_COST_FUNCTION = {
    'chi2': (CostFunction_Chi2, {}),
    'chi_2': (CostFunction_Chi2, {}),
    'chisquared': (CostFunction_Chi2, {}),
    'chi_squared': (CostFunction_Chi2, {}),
    'chi2_no_errors': (CostFunction_Chi2, {"errors_to_use": None, "add_determinant_cost": False}),
    'chi2_pointwise': (CostFunction_Chi2, {"errors_to_use": "pointwise"}),
    'chi2_pointwise_errors': (CostFunction_Chi2, {"errors_to_use": "pointwise"}),
    'chi2_covariance': (CostFunction_Chi2, {"errors_to_use": "covariance"}),
    'nll': (CostFunction_NegLogLikelihood, {"ratio": False}),
    'poisson': (CostFunction_NegLogLikelihood,
                {"data_point_distribution": "poisson", "ratio": False}),
    'nll-poisson': (CostFunction_NegLogLikelihood,
                    {"data_point_distribution": "poisson", "ratio": False}),
    'nll_poisson': (CostFunction_NegLogLikelihood,
                    {"data_point_distribution": "poisson", "ratio": False}),
    'nllpoisson': (CostFunction_NegLogLikelihood,
                    {"data_point_distribution": "poisson", "ratio": False}),
    'nll-gaussian': (CostFunction_NegLogLikelihood,
                    {"data_point_distribution": "gaussian", "ratio": False}),
    'nll_gaussian': (CostFunction_NegLogLikelihood,
                    {"data_point_distribution": "gaussian", "ratio": False}),
    'nllgaussiann': (CostFunction_NegLogLikelihood,
                    {"data_point_distribution": "gaussian", "ratio": False}),
    'negloglikelihood': (CostFunction_NegLogLikelihood, {"ratio": False}),
    'neg_log_likelihood': (CostFunction_NegLogLikelihood, {"ratio": False}),
    'nllr': (CostFunction_NegLogLikelihood, {"ratio": True}),
    'nllr-poisson': (CostFunction_NegLogLikelihood,
                    {"data_point_distribution": "poisson", "ratio": True}),
    'nllr_poisson': (CostFunction_NegLogLikelihood,
                    {"data_point_distribution": "poisson", "ratio": True}),
    'nllrpoisson': (CostFunction_NegLogLikelihood,
                    {"data_point_distribution": "poisson", "ratio": True}),
    'nllr-gaussian': (CostFunction_NegLogLikelihood,
                    {"data_point_distribution": "gaussian", "ratio": True}),
    'nllr_gaussian': (CostFunction_NegLogLikelihood,
                    {"data_point_distribution": "gaussian", "ratio": True}),
    'nllrgaussian': (CostFunction_NegLogLikelihood,
                    {"data_point_distribution": "gaussian", "ratio": True}),
    'negloglikelihoodratio': (CostFunction_NegLogLikelihood, {"ratio": True}),
    'neg_log_likelihood_ratio': (CostFunction_NegLogLikelihood, {"ratio": True}),
    'gauss-approximation': (CostFunction_GaussApproximation, {}),
    'gauss_approximation': (CostFunction_GaussApproximation, {}),
    'gauss_approximation_covariance': (
        CostFunction_GaussApproximation, {"errors_to_use": "covariance"}),
    'gauss_approximation_pointwise': (
        CostFunction_GaussApproximation, {"errors_to_use": "pointwise"}),
    'gauss_approximation_pointwise_errors': (
        CostFunction_GaussApproximation, {"errors_to_use": "pointwise"}),
}
